"""
Gry: the GRist pYthon command line, powered by Pygrister.
=========================================================

Gry is a command line tool for interacting with the Grist API. 
Gry is based on the Pygrister library and maps almost all the Grist APIs, 
even if offering sometimes a simplified, easier to type syntax. 

Gry shares the Pygrister configuration system, meaning that, if you already 
have working Pygrister config files/env variables in place, then Gry will 
also work. In addition, you may put a ``gryconf.json`` file in the current 
directory, and Gry (but not Pygrister) will look there too. Read the 
documentation for more info about the Gry/Pygrister configuration system. 

Gry is organized in "commands", *nouns* referring to the various sections 
of the Grist API: ``gry doc``, ``gry table``, ``gry team`` and so on.
Each command has "sub-commands", actions to perform (mostly *verbs*). 
Sub-commands tend to repeat across commands: ``gry doc new`` adds 
a document, ``gry table new`` adds a table, ``gry col new`` adds a column...

Type ``% gry --help`` to get a list of available commands. Then type, 
for instance, ``% gry doc --help`` to list the sub-commands available 
for ``gry doc``. Finally, something like ``% gry doc new --help`` will show 
how to use a particular sub-command.

Default output is human-readable; pass ``-v`` to get the original Pygrister 
output instead; pass ``-vv`` to get the underlying Grist response (as parsable 
json). Add ``-i`` to also get a dump of request metadata, for inspection. 
Pass ``-q`` to suppress all output. 

Use ``% gry python`` to open a special Python shell, pre-loaded with a 
working Pygrister instance: this way, you can alternate between Gry and 
Pygrister in the same session when needed. 

Example usage::

    % gry team see # get info on the "default" team as per config
    % gry doc see  # the "default" document as per config
    % gry doc see -d f4Y8Tov7TRkTQfUuj7TVdh # select a specific document
    # the best way to switch to another document, from now on: 
    % export GRIST_DOC_ID=f4Y8Tov7TRkTQfUuj7TVdh # or "set" in windows
    % gry doc see # the same as above, but no need to add the "-d" option
    % gry doc see -d bogus_doc # now this will fail...
    % gry doc see -d bogus_doc -i # ...so let's see the request details 
    % gry ws see -w 42 # workspace info, in a nicely formatted table
    % gry ws see -w 42 -vv # the same, in the original raw json
    % gry table new --help # how do I add a table?
    % gry table new name:Text:Name age:Int:Age --table People # like this!
    % gry col list -b People # the columns of our new table
    % gry rec new name:"John Doe" age:42 -b People # populate the table
    % gry sql "select * from People where age>?" -p 35 # run an sql query
    % gry python # let's open a Python shell now!
    >>> gry.list_cols(table_id='People') # "gry" is now a python object
    >>> exit() # and we are back to the shell

`Full documentation <https://pygrister.readthedocs.io>`_ 
`Grist API reference documentation <https://support.getgrist.com/api/>`_
"""

import os
import subprocess
import json as modjson
from sys import executable
from pathlib import Path
from enum import Enum
from typing import Any, List, Optional
from typing_extensions import Annotated

import typer
from rich.console import Console
from rich.table import Table
from requests import RequestException, ConnectionError, Timeout

from pygrister import __version__
from pygrister.api import GristApi
from pygrister.apicaller import ApiCaller
from pygrister.config import Configurator, apikey2output, PYGRISTER_CONFIG
try:
    from cliconverters import cli_out_converters # type: ignore
except (ModuleNotFoundError, ImportError):
    cli_out_converters = dict()
try:
    from cliconverters import cli_in_converters # type: ignore
except (ModuleNotFoundError, ImportError):
    cli_in_converters = dict()

# we borrow from Cloudflare non-standard 5xx codes 
# to express connection errors (other than HttpError) in Gry
CONN_ERROR = 523 # Requests ConnectionError -> Cf "Origin Is Unreachable"
CONN_TIMEOUT = 522 # Requests Timeout -> Cf "Connection Timed Out"
CONN_MISC_ERROR = 520 # other Requests Exceptions -> Cf "Web Server Returned an Unknown Error"

class _CliApiCaller(ApiCaller):
    """A custom api caller intended for the Pygrister cli."""
    def apicall(self, *args, **kwargs):
        # we try to catch possible connection errors, 
        # so that Gry won't die in flames in front of the user
        try:
            st, res = super().apicall(*args, **kwargs)
        except Timeout as e:
            st = CONN_TIMEOUT
            res = str(e)
        except ConnectionError as e:
            st = CONN_ERROR
            res = str(e)
        except RequestException as e:
            st = CONN_MISC_ERROR
            res = str(e)
        return st, res

class _CliConfigurator(Configurator):
    """A custom configurator intended for the Pygrister cli. 
    
    After loading the configuration from the usual place, 
    this will search for a (optional) "gryconf.json" file   
    located in the current directory, *before* looking for env vars. 
    Note: this configurator also overrides the ``GRIST_RAISE_ERROR``  
    and ``GRIST_SAFEMODE`` config keys, setting both to ``N``.
    """
    @staticmethod
    def get_config() -> dict[str, str]:
        config = dict(PYGRISTER_CONFIG)
        # default value for an additional config key
        config.update({'GRIST_GRY_TIMEOUT': '60'})
        pth = Path('~/.gristapi/config.json').expanduser()
        if pth.is_file():
            with open(pth, 'r', encoding='utf8') as f:
                config.update(modjson.loads(f.read()))
        pth = Path('gryconf.json')
        if pth.is_file(): 
            with open(pth, 'r', encoding='utf8') as f:
                config.update(modjson.loads(f.read()))
        for k in config.keys():
            try:
                config[k] = os.environ[k]
            except KeyError:
                pass
        # overrides
        config['GRIST_RAISE_ERROR'] = 'N'
        config['GRIST_SAFEMODE'] = 'N'
        return config

# the global GristApi (re-creadted at every cli call): inside a cli function, 
# all api calls (usually just one but may be more) will use this instance
_c = _CliConfigurator()
req_options = {'timeout': int(_c.config['GRIST_GRY_TIMEOUT'])}
pth = Path('gryrequest.json') # optional, to pass even more options to Requests
if pth.is_file(): 
    with open(pth, 'r', encoding='utf8') as f:
        req_options.update(modjson.loads(f.read()))
_a = _CliApiCaller(configurator=_c, request_options=req_options)
grist_api = GristApi(custom_apicaller=_a)
grist_api.in_converter = cli_in_converters
grist_api.out_converter = cli_out_converters

# the global Rich console where everything should be printed
cli_console = Console()

BADCALL = 3 # the exit code we reserve for bad call errors (eg http 404)
DONEMSG = '[green]Done.[/green]'
ERRMSG = '[bold red]Error![/bold red]'

# a few helper functions
# ----------------------------------------------------------------------
def _exit_if_error(st: int, res: Any, quiet: bool, 
                   verbose: int, inspect: bool) -> None:
    if inspect and not quiet:
        cli_console.print(grist_api.inspect())
        cli_console.rule()
    if not grist_api.ok:
        if not quiet:
            if verbose < 2:
                cli_console.print(ERRMSG, 'Status:', st, res)
            else:
                cli_console.print(grist_api.apicaller.response_as_json())
        raise typer.Exit(BADCALL)

def _print_done_or_exit(st: int, res: Any, quiet: bool, 
                        verbose: int, inspect: bool) -> None:
    if inspect and not quiet:
        cli_console.print(grist_api.inspect())
        cli_console.rule()
    if grist_api.ok: 
        if not quiet:
            if verbose == 0:
                cli_console.print(DONEMSG)
            elif verbose == 1:
                cli_console.print(res)
            else:
                cli_console.print(grist_api.apicaller.response_as_json())
    else:
        cli_console.print(ERRMSG, 'Status:', st, res)
        raise typer.Exit(BADCALL)

def _print_output(content: Any, res: Any, quiet: bool, 
                  verbose: int, inspect: bool) -> None:
    if inspect and not quiet:
        cli_console.print(grist_api.inspect())
        cli_console.rule()
    if not quiet:
        # we print different things, depending on verbose level
        # note that we always have the cli output ready, even if we won't use it
        if verbose == 0: # the nicely formatted cli output (text)
            cli_console.print(content)
        elif verbose == 1: # the response from Pygrister (Python object)
            cli_console.print(res)
        else: # the original Grist api response (json)
            cli_console.print(grist_api.apicaller.response_as_json())

def _print_done_and_id(content: Any, res: Any, quiet: bool, 
                       verbose: int, inspect: bool) -> None:
    if not quiet:
        if inspect:
            cli_console.print(grist_api.inspect())
            cli_console.rule()
        if verbose == 0:
            cli_console.print(DONEMSG, 'Id:', content)
        elif verbose == 1:
            cli_console.print(res)
        else:
            cli_console.print(grist_api.apicaller.response_as_json())

def _make_user_table(response: list[dict]) -> Table:
    table = Table('id', 'name', 'email', 'access')
    for usr in response:
        table.add_row(str(usr['id']), usr['name'], usr['email'], 
                      str(usr['access']))
    return table

def _make_scim_user_data(user: dict, table: Table) -> Table:
    table.add_row('id', str(user['id']))
    table.add_row('name', user['userName'])
    table.add_row('display name', user['displayName'])
    emails = []
    for email in user['emails']:
        e = email['value']
        if email['primary']:
            e += ' (primary)'
        emails.append(e)
    table.add_row('email', '\n'.join(emails))
    table.add_section()
    return table

def _user_access_validate(value: str|None):
    legal = 'owners editors viewers members none'
    if value not in legal.split():
        raise typer.BadParameter(f'Access must be one of: {legal}')
    if value == 'none':
        value = None
    return value

def _user_max_access_validate(value: str|None):
    legal = 'owners editors viewers'
    if value not in legal.split():
        raise typer.BadParameter(f'Access must be one of: {legal}')
    if value == 'none':
        value = None
    return value

def _column_decl_validate(value: list[str]):
    res = []
    for item in value:
        try:
            id_, type_, name = item.split(':')
        except ValueError:
            raise typer.BadParameter('Column must be declared as "id:type:label"')
        res.append([id_, type_, name])
    return res

def _record_decl_validate(value: list[str]):
    res = []
    for item in value:
        try:
            k, v = item.split(':')
        except ValueError:
            raise typer.BadParameter(
                'Record must be declared as "col:value col:value ..."')
        res.append([k, v])
    return res

def _variadic_options_validate(value: list[str]):
    try:
        return dict(zip(*[iter([i.strip('--') for i in value])]*2, strict=True))
    except ValueError:
        raise typer.BadParameter('Improper use of extra option(s)')

def _upload_path_validate(value: Path):
    if not value.is_file():
        raise typer.BadParameter(f'File does not exist: {value}')
    return value

def _upload_pathlist_validate(value: list[Path]):
    for item in value:
        _ = _upload_path_validate(item)
    return value

def _download_path_validate(value: Path):
    if not value.parent.is_dir():
        raise typer.BadParameter(f'Path does not exist: {value}')
    return value

# a few recurrent Typer options
# ----------------------------------------------------------------------
_verbose_opt = typer.Option('--verbose', '-v', count=True,
                            help='Verbose level (0-2)')
_quiet_opt = typer.Option('--quiet', '-q', help='All output will be suppressed')
_inspect_opt = typer.Option('--inspect', '-i', 
                            help = 'Print inspect output after api call')
_user_id_arg = typer.Argument(help='The user ID')
_team_id_opt = typer.Option('--team', '-t', 
                            help='The team ID [default: current]')
_ws_id_opt = typer.Option('--workspace', '-w', 
                          help='The workspace integer ID (0 means current)')
_doc_id_opt = typer.Option('--document', '-d', 
                           help='The document ID [default: current]')
_table_id_opt = typer.Option('--table', '-b',  help='The table ID name', 
                             prompt='Insert the table ID name') 
_access_opt = typer.Option('--access', '-a', 
                           help='The new access level',
                           prompt='Insert the new access level',
                           callback=_user_access_validate)
_max_access_opt = typer.Option('--max-access', '-A', 
                               help='The max inherited access level',
                               callback=_user_max_access_validate)
_pinned_opt = typer.Option('--pinned/--no-pinned', '-P/-p', help='Is pinned')
_hidden_opt = typer.Option('--hidden/--no-hidden', '-H/-h', 
                           help='Include hidden cols')
_limit_opt = typer.Option('--limit', '-l', 
                          help='Return at most this number of rows.')
_sort_opt = typer.Option('--sort', '-s', 
                         help='Order in which to return results.')
_noparse_opt = typer.Option('--noparse', 
                            help='True prohibits parsing according to col type')
_outmode_opt = typer.Option('--output-mode', '-m', help='Output type')
_enable_opt = typer.Option('--enable/--disable', help='Enable/disable')

# Typer sub-commands
# ----------------------------------------------------------------------
user_app = typer.Typer(help='Manage users, SCIM must be enabled')
org_app = typer.Typer(help='Manage Grist teams, aka organisations')
ws_app = typer.Typer(help='Manage workspaces inside a team site')
doc_app = typer.Typer(help='Manage documents inside a workspace')
table_app = typer.Typer(help='Manage tables inside a document')
col_app = typer.Typer(help='Manage columns inside a table')
rec_app = typer.Typer(help='Manage records inside a table')
att_app = typer.Typer(help='Manage attachments and attachment storage')
hook_app = typer.Typer(help='Manage document webhooks')
scim_app = typer.Typer(help='Metadata about SCIM services, if enabled')
_help = 'Gry, a command line tool for the Grist API - powered by Pygrister'
_epilog = 'Learn more: https://pygrister.readthedocs.io - https://github.com/ricpol/pygrister '
app = typer.Typer(no_args_is_help=True, help=_help, epilog=_epilog)
app.add_typer(user_app, name='user', no_args_is_help=True)
app.add_typer(org_app, name='team', no_args_is_help=True)
app.add_typer(ws_app, name='ws', no_args_is_help=True)
app.add_typer(doc_app, name='doc', no_args_is_help=True)
app.add_typer(table_app, name='table', no_args_is_help=True)
app.add_typer(col_app, name='col', no_args_is_help=True)
app.add_typer(rec_app, name='rec', no_args_is_help=True)
app.add_typer(att_app, name='att', no_args_is_help=True)
app.add_typer(hook_app, name='hook', no_args_is_help=True)
app.add_typer(scim_app, name='scim', no_args_is_help=True)

# gry version
# ----------------------------------------------------------------------
@app.command('version')
def gryversion() -> None:
    """Prints Pygrister/Gry version"""
    cli_console.print(__version__)

# gry test -> a quick configuration check
# ----------------------------------------------------------------------
@app.command('test')
def grytest() -> None:
    """Run a quick configuration test for your Gry console"""
    tests = {
        'connection': 'skipped', 
        'scim enabled': 'skipped',
        'default ws': 'skipped',
        'default team': 'skipped', 
        'default doc': 'skipped', 
        'store type': 'skipped',
        }
    abort = False
    content = Table('test', 'result')
    try:
        st, res = grist_api.see_team()
    except RequestException as e:
        tests['connection'] = str(e)
        abort = True
    if not abort and st == 401: # invalid api key
        tests['connection'] = str(res)
        abort = True # no sense in going on
    if not abort:
        tests['connection'] = 'ok'
        if st == 200:
            tests['default team'] = 'ok'
        else:
            tests['default team'] = str(res)
        st, res = grist_api.see_myself()
        tests['scim enabled'] = 'yes' if st == 200 else str(res)
        st, res = grist_api.see_workspace()
        tests['default ws'] = 'ok' if st == 200 else str(res)
        if tests['default team'] == 'ok':
            st, res = grist_api.see_doc()
            tests['default doc'] = 'ok' if st == 200 else str(res)
        if tests['default doc'] == 'ok':
            st, res = grist_api.see_attachment_store()
            tests['store type'] = str(res)
    for key, value in tests.items():
        content.add_row(key, value)
    _print_output(content, '', False, 0, False)

# gry conf -> print the current Grist configuration
# ----------------------------------------------------------------------
@app.command('conf')
def gryconf(
    showkey: Annotated[bool, 
                       typer.Option('--show-apikey/--hide-apikey', '-K/-k', 
                       help='Show in full or obfuscate apikey')] = False,
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0) -> None:
    """Print current Gry configuration and additional Requests options"""
    res = grist_api.configurator.config
    if not showkey:
        res['GRIST_API_KEY'] = apikey2output(res['GRIST_API_KEY'])
    table = Table('key', 'value')
    for k, v in res.items():
        table.add_row(k, str(v))
    table.add_section()
    for k, v in grist_api.apicaller.request_options.items():
        table.add_row(k, str(v))
    # _print_output is not quite made for this use case!
    if not quiet:
        if verbose == 0:
            cli_console.print(table)
        else:
            cli_console.print(res)

# gry sql -> post SELECT sql queries to Grist
# ----------------------------------------------------------------------
@app.command('sql')
def run_sql(
    statement: Annotated[str, typer.Argument(
                         help='The sql statement - SELECT only')],
    params: Annotated[Optional[List[str]], typer.Option('--param', '-p',
                      help='Query parameters')] = None,
    timeout: Annotated[int, typer.Option('--timeout', '-t', 
                       help='Query timeout')] = 1000,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Run a SELECT sql query against the document. 
    
    Repeat PARAM for multiple parameters, eg:
    
    % gry sql "select (...) where x>? and x<?;" -p 10 -p 100"""
    if params:
        st, res = grist_api.run_sql_with_args(statement, params, timeout, 
                                              doc_id, team_id)
    else:
        st, res = grist_api.run_sql(statement, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    if res:
        content = Table(*res[0].keys())
        for row in res:
            content.add_row(*[str(v) for v in row.values()])
    else:
        content = 'No records found.'
    _print_output(content, res, quiet, verbose, inspect)

# gry python -> open a Gry-aware Python shell
# ----------------------------------------------------------------------
@app.command('python')
def open_python(idle: Annotated[bool, typer.Option('--idle', 
                                help='Open the Idle shell')] = False):
    """Open a Python shell with Pygrister pre-loaded"""
    start = Path(__file__).absolute().parent / '_pygrystart.py'
    oldstartup = os.environ.get('PYTHONSTARTUP', None)
    os.environ['PYTHONSTARTUP'] = str(start)
    if idle:
        to_run = [executable, '-m', 'idlelib', '-s']
    else:
        to_run = [executable, '-q']
    try:
        subprocess.run(to_run)
    finally:
        if oldstartup:
            os.environ['PYTHONSTARTUP'] = oldstartup
        print('Done. Welcome back to Gry.')

# gry user -> for managing users with SCIM apis
# ----------------------------------------------------------------------
@user_app.command('list')
def list_users(
    start: Annotated[int, typer.Option('--start', '-s', 
                     help='First ID to retrieve')] = 1,
    retrieve: Annotated[int, typer.Option('--retrieve', '-r', 
                        help='Max users to retrieve')] = 10,
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List users. No filter option available"""
    st, res = grist_api.list_users_raw(start, retrieve)
    _exit_if_error(st, res, quiet, verbose, inspect)
    if start > res['totalResults']:
        content = 'No users.'
    else:
        content = Table('key', 'value')
        for user in res['Resources']:
            content = _make_scim_user_data(user, content)
    _print_output(content, res, quiet, verbose, inspect)

@user_app.command('me')
def see_me(quiet: Annotated[bool, _quiet_opt] = False,
           verbose: Annotated[int, _verbose_opt] = 0,
           inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Retrieve information about oneself"""
    st, res = grist_api.see_myself()
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('key', 'value')
    content = _make_scim_user_data(res, content)
    _print_output(content, res, quiet, verbose, inspect)

@user_app.command('see')
def see_user(user: Annotated[int, _user_id_arg],
             quiet: Annotated[bool, _quiet_opt] = False,
             verbose: Annotated[int, _verbose_opt] = 0,
             inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Retrieve user by ID"""
    st, res = grist_api.see_user(user)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('key', 'value')
    content = _make_scim_user_data(res, content)
    _print_output(content, res, quiet, verbose, inspect)
    
@user_app.command('new')
def new_user(
    name: Annotated[str, typer.Argument(help="User name")],
    email: Annotated[str, typer.Argument(help='User email')],
    display: Annotated[str, typer.Option('--display', '-d', 
                       help='User display name')] = '',
    formatted: Annotated[str, typer.Option('--formatted', '-f', 
                         help='User formatted name')] = '',
    lang: Annotated[str, typer.Option('--language', '-g', 
                    help='User language')] = 'en', 
    locale: Annotated[str, typer.Option('--locale', '-l', 
                      help='User locale')] = 'en',
    picture: Annotated[str, typer.Option('--picture', '-p', 
                       help='User picture url')] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Add a user"""
    photos = [picture] if picture else None
    st, res = grist_api.add_user(name, [email], formatted, display, lang, 
                                 locale, photos) 
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

class _OperationTypes(str, Enum):
    add = 'add'
    repl = 'replace'
    remove = 'remove'

@user_app.command('update')
def update_user(
    user_id: Annotated[int, typer.Argument(help="User ID")],
    op_path: Annotated[str, typer.Argument(help="Operation path")],
    op_value: Annotated[str, typer.Argument(help="Operation value")],
    operation: Annotated[_OperationTypes, typer.Option('--operation', '-o',         
                         help="Operation to perform")] = _OperationTypes.repl,
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Update a user. Only one update operation is possible"""
    op = {'op': operation, 'path': op_path, 'value': op_value}
    st, res = grist_api.update_user(user_id, [op])
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@user_app.command('delete')
def delete_user(user: Annotated[int, _user_id_arg],
                quiet: Annotated[bool, _quiet_opt] = False,
                verbose: Annotated[int, _verbose_opt] = 0,
                inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Remove a user"""
    st, res = grist_api.delete_user(user)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@user_app.command('enable')
def enable_user(user: Annotated[int, _user_id_arg],
                enable: Annotated[bool, _enable_opt] = True,
                quiet: Annotated[bool, _quiet_opt] = False,
                verbose: Annotated[int, _verbose_opt] = 0,
                inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Enable or disable a user"""
    st, res = grist_api.enable_user(user, enable)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

# TODO we don't implement "search" for now because we don't have a good way 
# to express filters in the shell. 

# gry scim -> metadata about SCIM service
# ----------------------------------------------------------------------
@scim_app.command('schemas')
def scim_schemas(quiet: Annotated[bool, _quiet_opt] = False,
                 verbose: Annotated[int, _verbose_opt] = 0,
                 inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Retrieve SCIM schemas"""
    st, res = grist_api.see_scim_schemas()
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

@scim_app.command('config')
def scim_config(quiet: Annotated[bool, _quiet_opt] = False,
                verbose: Annotated[int, _verbose_opt] = 0,
                inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Retrieve SCIM configuration"""
    st, res = grist_api.see_scim_config()
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

@scim_app.command('resources')
def scim_resources(quiet: Annotated[bool, _quiet_opt] = False,
                   verbose: Annotated[int, _verbose_opt] = 0,
                   inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Retrieve SCIM resources"""
    st, res = grist_api.see_scim_resources()
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

# gry team -> for managing team sites (organisations)
# ----------------------------------------------------------------------
@org_app.command('list')
def list_orgs(quiet: Annotated[bool, _quiet_opt] = False,
              verbose: Annotated[int, _verbose_opt] = 0,
              inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List the teams you have access to"""
    st, res = grist_api.list_team_sites()
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('id', 'name', 'owner')
    for org in res:
        try:
            owner = org['owner']['name']
        except TypeError:
            owner = 'Null'
        content.add_row(str(org['id']), org['name'], owner)
    _print_output(content, res, quiet, verbose, inspect)

@org_app.command('see')
def see_org(team_id: Annotated[str, _team_id_opt] = '', 
            quiet: Annotated[bool, _quiet_opt] = False,
            verbose: Annotated[int, _verbose_opt] = 0,
            inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Describe a team"""
    st, res = grist_api.see_team(team_id)
    _exit_if_error(st, res, quiet, verbose, inspect) 
    content = Table('key', 'value')
    content.add_row('id', str(res['id']))
    content.add_row('name', res['name'])
    content.add_row('domain', res['domain'])
    try:
        row = f"{res['owner']['id']} - {res['owner']['name']}"
    except TypeError:
        row = 'None'
    content.add_row('owner', row)
    _print_output(content, res, quiet, verbose, inspect)

@org_app.command('update')
def update_org(name: Annotated[str, typer.Argument(help='The new name')], 
               team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Change the team name"""
    st, res = grist_api.update_team(name, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)
     
@org_app.command('delete')
def delete_org(team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete a team"""
    st, res = grist_api.delete_team(team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@org_app.command('users')
def list_org_users(team_id: Annotated[str, _team_id_opt] = '',  
                   quiet: Annotated[bool, _quiet_opt] = False,
                   verbose: Annotated[int, _verbose_opt] = 0,
                   inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List users with access to team"""
    st, res = grist_api.list_team_users(team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = _make_user_table(res)
    _print_output(content, res, quiet, verbose, inspect)

@org_app.command('user-access')
def change_team_access(uid: Annotated[int, typer.Argument(help='The user ID')], 
                       access: Annotated[str, _access_opt],
                       team_id: Annotated[str, _team_id_opt] = '',
                       quiet: Annotated[bool, _quiet_opt] = False,
                       verbose: Annotated[int, _verbose_opt] = 0,
                       inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Change user access level to a team"""
    st, res = grist_api.list_team_users(team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    usr_email = ''
    for usr in res:
        if usr['id'] == uid:
            usr_email = usr['email']
            break
    if not usr_email:
        if inspect:
            cli_console.print(grist_api.inspect())
            cli_console.rule()
        raise typer.BadParameter('User id not found.')
    users = {usr_email: access}
    st, res = grist_api.update_team_users(users, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)
        
#TODO "change_user_access" is only for existing users, but "update_team_users"
# allows for adding users too. Maybe add a separate cli endpoint for this?

# gry ws -> for managing workspaces
# ----------------------------------------------------------------------
@ws_app.command('list')
def list_ws(team_id: Annotated[str, _team_id_opt] = '', 
            quiet: Annotated[bool, _quiet_opt] = False,
            verbose: Annotated[int, _verbose_opt] = 0,
            inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List workspaces and documents in a team"""
    st, res = grist_api.list_workspaces(team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('id', 'name', 'owner', 'email', 'docs')
    for ws in res:
        try:
            owner = str(ws['owner']['id'])
        except TypeError:
            owner = 'Null'
        try:
            email = ws['owner']['email']
        except TypeError:
            email = ''
        numdocs = len(ws.get('docs', []))
        content.add_row(str(ws['id']), ws['name'], owner, email, str(numdocs))
    _print_output(content, res, quiet, verbose, inspect)

@ws_app.command('see')
def see_ws(ws_id: Annotated[int, _ws_id_opt] = 0,
           quiet: Annotated[bool, _quiet_opt] = False,
           verbose: Annotated[int, _verbose_opt] = 0,
           inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Describe a workspace"""
    st, res = grist_api.see_workspace(ws_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('key', 'value')
    content.add_row('id', str(res['id']))
    content.add_row('name', res['name'])
    try:
        row = f"{res['org']['id']} - {res['org']['name']}"
    except TypeError:
        row = 'Null'
    content.add_row('team', row)
    content.add_section()
    for doc in res.get('docs', []):
        content.add_row('doc', f"{doc['id']} - {doc['name']}")
    _print_output(content, res, quiet, verbose, inspect)

@ws_app.command('new')
def add_ws(name: Annotated[str, typer.Argument(help='The name of new workspace')], 
           team_id: Annotated[str, _team_id_opt] = '',
           quiet: Annotated[bool, _quiet_opt] = False,
           verbose: Annotated[int, _verbose_opt] = 0,
           inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Create an empty workspace"""
    st, res = grist_api.add_workspace(name, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

@ws_app.command('update')
def update_ws(name: Annotated[str, typer.Argument(help='The new name')], 
              ws_id: Annotated[int, _ws_id_opt] = 0,
              quiet: Annotated[bool, _quiet_opt] = False,
              verbose: Annotated[int, _verbose_opt] = 0,
              inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Change the workspace name"""
    st, res = grist_api.update_workspace(name, ws_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@ws_app.command('delete')
def delete_ws(ws_id: Annotated[int, _ws_id_opt] = 0,
              quiet: Annotated[bool, _quiet_opt] = False,
              verbose: Annotated[int, _verbose_opt] = 0,
              inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete a workspace"""
    st, res = grist_api.delete_workspace(ws_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@ws_app.command('users')
def list_ws_users(ws_id: Annotated[int, _ws_id_opt] = 0,  
                  quiet: Annotated[bool, _quiet_opt] = False,
                  verbose: Annotated[int, _verbose_opt] = 0,
                  inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List users with access to workspace"""
    st, res = grist_api.list_workspace_users(ws_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = _make_user_table(res)
    _print_output(content, res, quiet, verbose, inspect)

@ws_app.command('user-access')
def change_ws_access(uid: Annotated[int, typer.Argument(help='The user ID')], 
                     access: Annotated[str, _access_opt],
                     ws_id: Annotated[int, _ws_id_opt] = 0,
                     quiet: Annotated[bool, _quiet_opt] = False,
                     verbose: Annotated[int, _verbose_opt] = 0,
                     inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Change user access level to a workspace"""
    st, res = grist_api.list_workspace_users(ws_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    usr_email = ''
    for usr in res:
        if usr['id'] == uid:
            usr_email = usr['email']
            break
    if not usr_email:
        if inspect:
            cli_console.print(grist_api.inspect())
            cli_console.rule()
        raise typer.BadParameter('User id not found.')
    users = {usr_email: access}
    st, res = grist_api.update_workspace_users(users, ws_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)
        
#TODO "change_ws_access" is only for existing users, but "update_team_users"
# allows for adding users too. Maybe add a separate cli endpoint for this?

# gry doc -> for managing documents
# ----------------------------------------------------------------------
@doc_app.command('see')
def see_doc(doc_id: Annotated[str, _doc_id_opt] = '', 
            team_id: Annotated[str, _team_id_opt] = '',
            quiet: Annotated[bool, _quiet_opt] = False,
            verbose: Annotated[int, _verbose_opt] = 0,
            inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Describe a document"""
    st, res = grist_api.see_doc(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('key', 'value')
    content.add_row('id', res['id'])
    content.add_row('name', res['name'])
    content.add_row('pinned', str(res['isPinned']))
    try:
        row = f"{res['workspace']['id']} - {res['workspace']['name']}"
    except TypeError:
        row = 'Null'
    content.add_row('workspace', row)
    try:
        row = f"{res['workspace']['org']['id']} - {res['workspace']['org']['name']}"
    except TypeError:
        row = 'Null'
    content.add_row('team', row)
    _print_output(content, res, quiet, verbose, inspect)

@doc_app.command('new')
def add_doc(name: Annotated[str, typer.Argument(help='The name of new doc')], 
            pinned: Annotated[bool, _pinned_opt] = False,
            ws_id: Annotated[int, _ws_id_opt] = 0, 
            quiet: Annotated[bool, _quiet_opt] = False,
            verbose: Annotated[int, _verbose_opt] = 0,
            inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Create an empty document"""
    st, res = grist_api.add_doc(name, pinned, ws_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

@doc_app.command('update')
def update_doc(name: Annotated[str, typer.Argument(help='The new name')],
               pinned: Annotated[bool, _pinned_opt] = False,
               doc_id: Annotated[str, _doc_id_opt] = '', 
               team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Modify document metadata"""
    st, res = grist_api.update_doc(name, pinned, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@ doc_app.command('move')
def move_doc(dest: Annotated[int, typer.Argument(help='Destination workspace ID')],
             doc_id: Annotated[str, _doc_id_opt] = '', 
             team_id: Annotated[str, _team_id_opt] = '',
             quiet: Annotated[bool, _quiet_opt] = False,
             verbose: Annotated[int, _verbose_opt] = 0,
             inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Move document to another workspace"""
    st, res = grist_api.move_doc(dest, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@ doc_app.command('copy')
def copy_doc(
    dest: Annotated[int, typer.Argument(help='Destination workspace ID')],
    name: Annotated[str, typer.Argument(help='Target doc name')],
    template: Annotated[bool,typer.Option('--template', 
                        help='Copy as template')] = False,             
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Copy document to another workspace"""
    st, res = grist_api.copy_doc(dest, name, template, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

@doc_app.command('delete')
def delete_doc(doc_id: Annotated[str, _doc_id_opt] = '', 
               team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete a document"""
    st, res = grist_api.delete_doc(doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@doc_app.command('purge-history')
def delete_doc_history(
    keep: Annotated[int, typer.Option('--keep', '-k', 
                    help='Latest actions to keep')] = 0,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete the document action history"""
    st, res = grist_api.delete_doc_history(keep, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@doc_app.command('reload')
def reload_doc(doc_id: Annotated[str, _doc_id_opt] = '', 
               team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Reload a document"""
    st, res = grist_api.reload_doc(doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@doc_app.command('enable')
def enable_doc(enable: Annotated[bool, _enable_opt] = True,
               doc_id: Annotated[str, _doc_id_opt] = '', 
               team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Ebable or disable a document"""
    st, res = grist_api.enable_doc(enable, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@doc_app.command('recovery')
def doc_recovery(
    mode: Annotated[bool, typer.Option('--set/--unset', '-R/-r', 
                    help='Set/unset recovery mode')] = True,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Controls the recovery mode of a document"""
    st, res = grist_api.set_recovery_mode(mode, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@doc_app.command('download')
def download_db(
    filename: Annotated[Path, typer.Argument(help='Output file path',
                        callback=_download_path_validate)],
    history: Annotated[bool, typer.Option('--history/--no-history', '-H/-h', 
                       help='Include history')] = False, 
    template: Annotated[bool, typer.Option('--template/--no-template', '-P/-p', 
                        help='Template only, no data')] = False, 
    doc_id: Annotated[str, _doc_id_opt] = '', 
    quiet: Annotated[bool, _quiet_opt] = False,
    team_id: Annotated[str, _team_id_opt] = '',
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Download the document as sqlite file"""
    nohistory = not history
    st, res = grist_api.download_sqlite(filename, nohistory, template, 
                                        doc_id, team_id)
    _print_done_or_exit(st, res, quiet, 0, inspect) # force verbose=0 in download mode

@doc_app.command('upload')
def upload_doc(
    filename: Annotated[Path, typer.Argument(help='Sqlite file to upload', 
                        callback=_upload_path_validate)],
    docname: Annotated[str, typer.Argument(help='Name of the target document')],
    ws_id: Annotated[int, _ws_id_opt] = 0,
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Import a document from an sqlite file"""
    st, res = grist_api.upload_sqlite(filename, docname, ws_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

@doc_app.command('users')
def list_doc_users(doc_id: Annotated[str, _doc_id_opt] = '', 
                   team_id: Annotated[str, _team_id_opt] = '',
                   quiet: Annotated[bool, _quiet_opt] = False,
                   verbose: Annotated[int, _verbose_opt] = 0,
                   inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List users with access to document"""
    st, res = grist_api.list_doc_users(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = _make_user_table(res)
    _print_output(content, res, quiet, verbose, inspect)

@doc_app.command('user-access')
def change_doc_access(uid: Annotated[int, typer.Argument(help='The user ID')], 
                      access: Annotated[str, _access_opt],
                      max_access: Annotated[str, _max_access_opt] = 'owners', 
                      doc_id: Annotated[str, _doc_id_opt] = '', 
                      team_id: Annotated[str, _team_id_opt] = '',
                      quiet: Annotated[bool, _quiet_opt] = False,
                      verbose: Annotated[int, _verbose_opt] = 0,
                      inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Change user access level to a document"""
    st, res = grist_api.list_doc_users(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    usr_email = ''
    for usr in res:
        if usr['id'] == uid:
            usr_email = usr['email']
            break
    if not usr_email:
        if inspect:
            cli_console.print(grist_api.inspect())
            cli_console.rule()
        raise typer.BadParameter('User id not found.')
    users = {usr_email: access}
    st, res = grist_api.update_doc_users(users, max_access, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

#TODO again, this is for existing users only, but the api 
# allows for adding users too. Maybe add a separate cli endpoint for this?

# gry table -> for managing tables
# ----------------------------------------------------------------------
@table_app.command('list')
def list_tables(doc_id: Annotated[str, _doc_id_opt] = '', 
                team_id: Annotated[str, _team_id_opt] = '',
                quiet: Annotated[bool, _quiet_opt] = False,
                verbose: Annotated[int, _verbose_opt] = 0,
                inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List tables in a document"""
    st, res = grist_api.list_tables(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('table id', 'metadata')
    for t in res:
        f = t['fields']
        mdata = '\n'.join([f'{k}: {v}' for k, v in f.items()])
        content.add_row(t['id'], mdata)
        content.add_section()
    _print_output(content, res, quiet, verbose, inspect)

@table_app.command('new')
def new_table(
    cols: Annotated[List[str], typer.Argument(callback=_column_decl_validate,
                    help='Column list, each declared as "id:type:label"')],
    tname: Annotated[str, _table_id_opt], 
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Add one table to a document. 

    Column must be declared as "id:type:label"; type is any valid Grist type:

    % gry table new name:Text:Name age:Int:Age -n People"""
    column_list = []
    for id_, type_, label in cols:
        column_list.append({'id': id_, 
                            'fields': {'type': type_, 'label': label}})
    table = [{'id': tname, 'columns': column_list}]
    st, res = grist_api.add_tables(table, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res[0], res, quiet, verbose, inspect)

@table_app.command('update', context_settings={'allow_extra_args': True, 
                                               'ignore_unknown_options': True})
def update_table(ctx: typer.Context,
                 tname: Annotated[str, _table_id_opt], 
                 doc_id: Annotated[str, _doc_id_opt] = '', 
                 team_id: Annotated[str, _team_id_opt] = '',
                 quiet: Annotated[bool, _quiet_opt] = False,
                 verbose: Annotated[int, _verbose_opt] = 0,
                 inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Update table metadata.
    
    Pass any metadata field as an extra option, eg: 
    
    % gry table update -b Mytable --tableRef 2 --onDemand false"""
    fields = _variadic_options_validate(ctx.args)
    # some more validation 
    valid_keys = ('primaryViewId', 'summarySourceTable', 'onDemand', 
                  'rawViewSectionRef', 'recordCardViewSectionRef')
    for k in fields.keys():
        if k not in valid_keys:
            raise typer.BadParameter(f'Unknown option: {k}')
        if k == 'onDemand':
            if fields[k] not in ('true', 'false'):
                raise typer.BadParameter(f'onDemand option must be "true" or "false"')
        else:
            try:
                fields[k] = int(fields[k])
            except ValueError:
                raise typer.BadParameter(f'Option {k} value must be integer')
    table = [{'id': tname, 'fields': fields}]
    st, res = grist_api.update_tables(table, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

class _DownloadTableOption(str, Enum):
    excel = 'excel'
    csv = 'csv'
    schema = 'schema'

class _HeaderOption(str, Enum):
    label = 'label'
    colid = 'colId'

@table_app.command('download')
def download_table(
    filename: Annotated[Path, typer.Argument(help='Output file path', 
                        callback=_download_path_validate)],
    tname: Annotated[str, _table_id_opt], 
    output: Annotated[_DownloadTableOption, _outmode_opt] = _DownloadTableOption.csv,
    header: Annotated[_HeaderOption, typer.Option('--header', '-h', 
                      help='Column headers')] = _HeaderOption.label,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Dumps content or schema of a table to FILENAME"""
    funcs = {_DownloadTableOption.csv: grist_api.download_csv, 
             _DownloadTableOption.excel: grist_api.download_excel,
             _DownloadTableOption.schema: grist_api.download_excel}
    st, res = funcs[output](filename, tname, header, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, 0, inspect) # force verbose=0 in download mode

# gry col -> for managing columns
# ----------------------------------------------------------------------
@col_app.command('list')
def list_columns(
    table: Annotated[str, _table_id_opt],
    hidden: Annotated[bool, _hidden_opt] = False,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List columns in a table"""
    st, res = grist_api.list_cols(table, hidden, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('column id', 'metadata')
    for c in res:
        f = c['fields']
        useful_fields = ['label', 'type', 'isFormula', 'formula']
        mdata = '\n'.join([f'{i}: {str(f[i])}' for i in useful_fields ])
        content.add_row(c['id'], mdata)
        content.add_section()
    _print_output(content, res, quiet, verbose, inspect)

@col_app.command('new')
def add_column(
    cols: Annotated[List[str], typer.Argument(callback=_column_decl_validate,
                    help='Column list, each declared as "id:type:label"')],
    table: Annotated[str, _table_id_opt],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Add columns to a table.
    
    Column must be declared as "id:type:label"; type is any valid Grist type:

    % gry col new name:Text:Name age:Int:Age -b MyTable"""
    columns = []
    for id_, type_, label in cols:
        columns.append({'id': id_, 
                        'fields': {'type': type_, 'label': label}})
    st, res = grist_api.add_cols(table, columns, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

@col_app.command('update')
def update_column(
    cols: Annotated[List[str], typer.Argument(callback=_column_decl_validate,
                    help='Column list, each declared as "id:type:label"')],
    table: Annotated[str, _table_id_opt],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Update columns in a table.
    
    Column must be declared as "id:type:label"; type is any valid Grist type:

    % gry col update name:Text:Name age:Int:Age -b MyTable"""
    columns = []
    for id_, type_, label in cols:
        columns.append({'id': id_, 
                        'fields': {'type': type_, 'label': label}})
    st, res = grist_api.update_cols(table, columns, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

# Note: we don't cover add_update_cols (the PUT api) because it is too 
# sophisticated for our very limited way of describing a column in the cli

@col_app.command('delete')
def delete_column(
    col: Annotated[str, typer.Argument(help='Name ID of the column to delete')],
    table: Annotated[str, _table_id_opt],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete a column"""
    st, res = grist_api.delete_column(table, col, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

# gry rec -> for managing records
# ----------------------------------------------------------------------
@rec_app.command('list')
def list_records(table: Annotated[str, _table_id_opt],
                 sort: Annotated[str, _sort_opt] = '',
                 limit: Annotated[int, _limit_opt] = 0,
                 hidden: Annotated[bool, _hidden_opt] = False,
                 doc_id: Annotated[str, _doc_id_opt] = '', 
                 team_id: Annotated[str, _team_id_opt] = '',
                 quiet: Annotated[bool, _quiet_opt] = False,
                 verbose: Annotated[int, _verbose_opt] = 0,
                 inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Fetch records from a table.
    
    No filter option is available here: 'gry sql' may be a better try."""
    st, res = grist_api.list_records(table, sort=sort, limit=limit, 
                            hidden=hidden, doc_id=doc_id, team_id=team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    if res:
        content = Table(*res[0].keys())
        for row in res:
            content.add_row(*[str(v) for v in row.values()])
    else:
        content = 'No records found.'
    _print_output(content, res, quiet, verbose, inspect)
    
@rec_app.command('new')
def add_record(
    record: Annotated[List[str], typer.Argument(callback=_record_decl_validate,
                      help='One record, declared as "col:value col:value ..."')],
    table: Annotated[str, _table_id_opt],
    noparse: Annotated[bool, _noparse_opt] = False,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Add one record to a table"""
    record = dict(record) # type: ignore
    st, res = grist_api.add_records(table, [record], noparse, doc_id, team_id) # type: ignore
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res[0], res, quiet, verbose, inspect)

@rec_app.command('update')
def update_record(
    record: Annotated[List[str], typer.Argument(callback=_record_decl_validate,
                      help='One record, declared as "id:value col:value ..."')],
    table: Annotated[str, _table_id_opt],
    noparse: Annotated[bool, _noparse_opt] = False,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Modify one record of a table"""
    record = dict(record) # type: ignore
    try:
        record['id'] = int(record['id']) # type: ignore
    except ValueError:
        raise typer.BadParameter('Record ID must be a number')
    st, res = grist_api.update_records(table, [record], noparse, doc_id, team_id) # type: ignore
    _print_done_or_exit(st, res, quiet, verbose, inspect)

# Note: again, we don't cover add_update_records because it is too 
# sophisticated for our very limited way of describing a record in the cli

@rec_app.command('delete')
def delete_records(
    records: Annotated[List[int], typer.Argument(
                       help='ID of the records to delete')],
    table: Annotated[str, _table_id_opt],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete records from a table"""
    st, res = grist_api.delete_rows(table, records, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

# gry att -> for managing attachments
# ----------------------------------------------------------------------
@att_app.command('list')
def list_atts(sort: Annotated[str, _sort_opt] = '',
              limit: Annotated[int, _limit_opt] = 0,
              doc_id: Annotated[str, _doc_id_opt] = '', 
              team_id: Annotated[str, _team_id_opt] = '',
              quiet: Annotated[bool, _quiet_opt] = False,
              verbose: Annotated[int, _verbose_opt] = 0,
              inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List metadata of attachments in a doc. No filter option available"""
    st, res = grist_api.list_attachments(sort=sort, limit=limit, doc_id=doc_id, 
                                         team_id=team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('att. id', 'metadata')
    for c in res:
        f = c['fields']
        mdata = '\n'.join([f'{k}: {str(v)}' for k, v in f.items()])
        content.add_row(str(c['id']), mdata)
        content.add_section()
    _print_output(content, res, quiet, verbose, inspect)

@att_app.command('see')
def see_attachment(att_id: Annotated[int, typer.Argument(help='The attachment ID')],
                   doc_id: Annotated[str, _doc_id_opt] = '', 
                   team_id: Annotated[str, _team_id_opt] = '',
                   quiet: Annotated[bool, _quiet_opt] = False,
                   verbose: Annotated[int, _verbose_opt] = 0,
                   inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Get the metadata for an attachment"""
    st, res = grist_api.see_attachment(att_id, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    content = Table('key', 'value')
    content.add_row('ID', str(att_id))
    for k, v in res.items():
        content.add_row(k, str(v))
    _print_output(content, res, quiet, verbose, inspect)

@att_app.command('download')
def download_att(
    filename: Annotated[Path, typer.Argument(help='Output file path', 
                        callback=_download_path_validate)],
    attachment: Annotated[int, typer.Option('--attachment', '-a',
                          help='The attachment ID', 
                          prompt='Insert the attachment ID')],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Download one attachment as a file"""
    st, res = grist_api.download_attachment(filename, attachment, 
                                            doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)
    
@att_app.command('upload')
def upload_atts(
    filenames: Annotated[List[Path], typer.Argument(help='Files to upload', 
                         callback=_upload_pathlist_validate)],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Upload one or more attachments to a doc"""
    st, res = grist_api.upload_attachments(filenames, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_and_id(res, res, quiet, verbose, inspect)

class _DownloadAttOption(str, Enum):
    tar = 'tar'
    zip = 'zip'

@att_app.command('backup')
def download_atts(
    filename: Annotated[Path, typer.Argument(help='Output file path', 
                        callback=_download_path_validate)],
    output: Annotated[_DownloadAttOption, _outmode_opt] = _DownloadAttOption.tar,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Download all attachments as an archive file"""
    st, res = grist_api.download_attachments(filename, output, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@att_app.command('restore')
def upload_restore_atts(
    filename: Annotated[Path, typer.Argument(help='Archive file path', 
                        callback=_upload_path_validate)],
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Upload missing attachments from a local tar archive"""
    st, res = grist_api.upload_restore_attachments(filename, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

@att_app.command('purge')
def purge_atts(
    old: Annotated[bool, typer.Option('--only-old', 
                   help='Only if unused for a long time')] = False,  
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete unused attachments from the document"""
    st, res = grist_api.delete_unused_attachments(old, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@att_app.command('store')
def see_att_store(doc_id: Annotated[str, _doc_id_opt] = '', 
                  team_id: Annotated[str, _team_id_opt] = '',
                  quiet: Annotated[bool, _quiet_opt] = False,
                  verbose: Annotated[int, _verbose_opt] = 0,
                  inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Get the attachments storage type"""
    st, res = grist_api.see_attachment_store(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

@att_app.command('set-store')
def change_att_store(
    internal: Annotated[bool, typer.Option('--internal/--external', '-i/-e', 
                        help='The storage type')] = True,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Set the attachments storage type"""
    st, res = grist_api.update_attachment_store(internal, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@att_app.command('store-settings')
def list_store_settings(doc_id: Annotated[str, _doc_id_opt] = '', 
                        team_id: Annotated[str, _team_id_opt] = '',
                        quiet: Annotated[bool, _quiet_opt] = False,
                        verbose: Annotated[int, _verbose_opt] = 0,
                        inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Get the attachments storage settings"""
    st, res = grist_api.list_store_settings(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

@att_app.command('transfer')
def transfer_atts(doc_id: Annotated[str, _doc_id_opt] = '', 
                  team_id: Annotated[str, _team_id_opt] = '',
                  quiet: Annotated[bool, _quiet_opt] = False,
                  verbose: Annotated[int, _verbose_opt] = 0,
                  inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Start transferring attachments"""
    st, res = grist_api.transfer_attachments(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

@att_app.command('transfer-status')
def transfer_status(doc_id: Annotated[str, _doc_id_opt] = '', 
                    team_id: Annotated[str, _team_id_opt] = '',
                    quiet: Annotated[bool, _quiet_opt] = False,
                    verbose: Annotated[int, _verbose_opt] = 0,
                    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Get attachment transfer status"""
    st, res = grist_api.see_transfer_status(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res, res, quiet, verbose, inspect)

# gry hook -> for managing webhooks
# ----------------------------------------------------------------------
@hook_app.command('list')
def list_hooks(doc_id: Annotated[str, _doc_id_opt] = '', 
               team_id: Annotated[str, _team_id_opt] = '',
               quiet: Annotated[bool, _quiet_opt] = False,
               verbose: Annotated[int, _verbose_opt] = 0,
               inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """List webhooks associated with a document"""
    st, res = grist_api.list_webhooks(doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    if not res:
        content = 'No webhooks.'
    else:
        content = Table('webhook data')
        for wh in res:
            data = f'id: {wh["id"]}\n'
            data += f'name: {wh["fields"]["name"]}\n'
            data += f'url: {wh["fields"]["url"]}\n'
            data += f'enabled: {wh["fields"]["enabled"]}\n'
            data += f'table: {wh["fields"]["tableId"]}\n'
            data += f'events: {", ".join(wh["fields"]["eventTypes"])}'
            content.add_row(data)
            content.add_section()
    _print_output(content, res, quiet, verbose, inspect)

@hook_app.command('new')
def add_hook(
    name: Annotated[str, typer.Argument(help='Webhook name')],
    url: Annotated[str, typer.Argument(help='Webhook url')],
    table: Annotated[str, _table_id_opt],
    events: Annotated[str, typer.Option('--events',  
                      help='Event types :-separated, eg "add:update"')] = 'add',
    readycol: Annotated[str|None, typer.Option('--ready',  
                        help='Is Ready Columm (str or null)')] = None,
    enabled: Annotated[bool, typer.Option('--enabled/--disabled',  
                       help='Webhook is enabled')] = True,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Add one webhook to a document"""
    evts = events.split(':')
    wh = {'fields': {'name': name, 'memo': '', 'url': url, 'enabled': enabled, 
          'eventTypes': evts, 'isReadyColumn': readycol, 'tableId': table}}
    st, res = grist_api.add_webhooks([wh], doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_output(res[0], res, quiet, verbose, inspect)

@hook_app.command('update')
def update_hook(
    hook_id: Annotated[str, typer.Argument(help='The webhook ID')],
    name: Annotated[str, typer.Option('--name', help='Webhook name')] = '<same>',
    url: Annotated[str, typer.Option('--url', help='Webhook url')] = '<same>',
    table: Annotated[str, typer.Option('--table', help='Table ID name')] = '<same>',
    events: Annotated[str, typer.Option('--events', 
                      help='Event types :-separated, eg "add:update"')] = '<same>',
    readycol: Annotated[str|None, typer.Option('--ready',
                        help='Is Ready Columm (str or null)')] = '<same>',
    enabled: Annotated[bool|None, typer.Option('--enabled/--disabled',  
                       help='Webhook is enabled')] = None,
    doc_id: Annotated[str, _doc_id_opt] = '', 
    team_id: Annotated[str, _team_id_opt] = '',
    quiet: Annotated[bool, _quiet_opt] = False,
    verbose: Annotated[int, _verbose_opt] = 0,
    inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Modify a webhook"""
    fields = dict()
    for option, label in ((name, 'name'), (url, 'url'), (table, 'tableId'),
                          (readycol, 'isReadyColumn')):
        if option != '<same>':
            fields[label] = option
    if enabled is not None:
        fields['enabled'] = enabled
    if events != '<same>':
        evts = events.split(':')
        fields['eventTypes'] = evts
    wh = {'fields': fields}
    st, res = grist_api.update_webhook(hook_id, wh, doc_id, team_id)
    _exit_if_error(st, res, quiet, verbose, inspect)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@hook_app.command('delete')
def delete_hook(hook_id: Annotated[str, typer.Argument(help='The webhook ID')],
                doc_id: Annotated[str, _doc_id_opt] = '', 
                team_id: Annotated[str, _team_id_opt] = '',
                quiet: Annotated[bool, _quiet_opt] = False,
                verbose: Annotated[int, _verbose_opt] = 0,
                inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Delete a webhook"""
    st, res = grist_api.delete_webhook(hook_id, doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)

@hook_app.command('empty-queue')
def empty_hook_queue(doc_id: Annotated[str, _doc_id_opt] = '', 
                     team_id: Annotated[str, _team_id_opt] = '',
                     quiet: Annotated[bool, _quiet_opt] = False,
                     verbose: Annotated[int, _verbose_opt] = 0,
                     inspect: Annotated[bool, _inspect_opt] = False) -> None:
    """Empty a document's queue of undelivered payloads"""
    st, res = grist_api.empty_payloads_queue(doc_id, team_id)
    _print_done_or_exit(st, res, quiet, verbose, inspect)


if __name__ == '__main__':
    app()



# type: ignore
"""
This is the Pygrister test suite.
=================================

Before running the tests, you must set up a few things on the Grist side.
Either if you plan to run the tests against the standard SaaS Grist 
or against a self-managed instance of Grist, 
- find out or create your API key;
- create an empty team site - that is, only the "Home" workspace should 
  be present: in SaaS Grist, a free site will do;
  - (but, if you are running a self-managed "mono-site" Grist, then of course 
    the team site will be the one indicated by the ``GRIST_SINGLE_ORG`` 
    env variable that you will provide separately.)

If you are planning to run the tests many times against the standard SaaS Grist, 
in fact you may want to start with an entirely separate Grist account, to avoid 
cluttering your "home" db with bogus users data (see below).

Remember, the test suite *will not make use* of your regular configuration 
files (eg., ``~/.gristapi/config.json``). 
Everything *must* be provided in a separate ``config_test.json`` in the same 
directory as this file. Edit the already present json file as follows:
- "GRIST_API_KEY": your api key
- "GRIST_TEAM_SITE": your team site (you may ignore if running a mono-site Grist)
- "GRIST_SELF_MANAGED": set to "Y" if you are running a self-managed Grist
- "GRIST_SELF_MANAGED_HOME": the main url of your self-managed Grist
- "GRIST_SELF_MANAGED_SINGLE_ORG": "Y" if you are running the mono-team flavour 
- "GRIST_TEST_RUN_USER_TESTS": "N" to skip the tests that create users 
  (special, see below)
- "GRIST_TEST_RUN_SCIM_TESTS": "N" to skip all SCIM api tests (special, see below)
- "GRIST_TEST_RUN_EXT_ATTACH_TESTS": "N" to skip the tests that need an external 
  storage for attachments (special, see below)

Note: no other config key should be configured for the test suite. 

The test suite will leave several objects (workspaces and docs) in your 
team site. The objects created have unique names, so it should be safe 
re-running the test suite with the same sites: however, you may want 
to clean up eventually. 

Test users are a separate problem. A few tests will create new users in your 
test team site: even if you remove them manually in the "manage team" 
section of the GUI, you are only deleting the user-team association, but 
the user data will remain in your "home" database (which you currently cannot 
access in SaaS Grist). You really can't completely delete a user from the GUI 
and, in SaaS Grist, you can't use the SCIM apis (which are currently 
not enabled). It is normally not a problem to leave a few inactive users 
in the database. However, if you end up running the test suite hundreds of 
times, you may want to avoid the clutter. In a self-managed instance of Grist, 
you just throw the container away after a while. In SaaS Grist, you may want 
to activate a separate account to run the tests. 
To help you keep test users under control, we added a special configuration 
key, only useful when running the test suite: set "GRIST_TEST_RUN_USER_TESTS" 
to "N" to skip every test that may create a user. 

SCIM apis are a separate problem: SCIM tests are automatically skipped when you 
run the test suite against the regular Grist SaaS (because SCIM is not enabled 
there); but, even in self-hosted Grist, you have to enable SCIM first by 
setting the GRIST_ENABLE_SCIM env variable. If you are self-hosted, but you 
don't want to enable SCIM in your environment, set our special configuration key 
"GRIST_TEST_RUN_SCIM_TESTS" to "N" (in config_test.json file) to skip all SCIM 
tests altogether, so you don't get any spurious errors. 

Another separate problem is how to test the Apis that require an external storage 
for attachments. If your testing environment already supports external storage 
(that is, you have a bucket in place and a "GRIST_EXTERNAL_ATTACHMENTS_MODE" 
variable set to "snapshots") then you may set our special configuration key 
"GRIST_TEST_RUN_EXT_ATTACH_TESTS" to include the tests dealing with storages. 
If not, leave the key unset (the default) to skip those tests. 

A few object will also be downloaded in your current directory. Moreover, 
tests involving user/permission manipulation may trigger email notifications 
from Grist. 

Finally, please keep in mind that the purpose here is testing the Pygrister 
functions, not the underlying Grist api. Basically, we make sure that the 
api calls work as expected from our side of the business, but the calls 
themselves are always pretty basic. 

**Note**: if you are reading this file to learn how to use Pygrister, 
please note that all the function calls here always pass the "id params" 
(``team_id``, ``doc_id``...) explicitly. This is because we need to 
insulate the test suite from any external config file, but in real life 
Pygrister will pick up the missing parts from the configuration. 
So, whenever you see here a function call like 
``<grist>.add_records(<table>, records, doc_id=self.doc_id, team_id=self.team_id)``, 
in normal usage it's just ``<grist>.add_records(<table>, records)``.

"""

from pathlib import Path
import time
from datetime import datetime
import json
import unittest
from requests import RequestException, HTTPError, ConnectTimeout, PreparedRequest

from pygrister import api
from pygrister import config
from pygrister.apicaller import ApiCaller
from pygrister.exceptions import *

# standard config values used by tests
default_config = {
    "GRIST_SERVER_PROTOCOL": "https://",
    "GRIST_API_SERVER": "getgrist.com",
    "GRIST_API_ROOT": "api",
    "GRIST_RAISE_ERROR": "Y",
    "GRIST_SAFEMODE": "N",
    "GRIST_WORKSPACE_ID": "1000000000000", # no test should ever hit this
    "GRIST_DOC_ID": "_bogus_",             # no test should ever hit this
}

HERE = Path(__file__).absolute().parent
with open(HERE / 'config_test.json') as f:
    TEST_CONFIGURATION = json.loads(f.read())
TEST_CONFIGURATION.update(default_config)

total_apicalls = [] # behold, the hack of the mutable global!
def tearDownModule():
    print('\nTotal api calls:', sum(total_apicalls))


# two helper functions to prepare playground worspaces and documents
def _make_ws(team_id):
    gristapi = api.GristApi(config=TEST_CONFIGURATION)
    name = 'ws'+str(time.time_ns())
    st, ws_id = gristapi.add_workspace(name, team_id)
    total_apicalls.append(gristapi.apicaller.apicalls)
    return ws_id, name

def _make_doc(ws_id):
    gristapi = api.GristApi(config=TEST_CONFIGURATION)
    name = 'doc'+str(time.time_ns())
    st, doc_id = gristapi.add_doc(name, ws_id=ws_id)
    total_apicalls.append(gristapi.apicaller.apicalls)
    return doc_id


class BaseTestPyGrister(unittest.TestCase):
    def setUp(self):
        self.g = api.GristApi(config=TEST_CONFIGURATION)
    
    def tearDown(self):
        total_apicalls.append(self.g.apicaller.apicalls)


class TestVarious(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        cls.ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(cls.ws_id)

    def test_dry_run(self):
        self.g.apicaller.dry_run = True
        st, res = self.g.see_team()
        self.assertEqual(st, 418)
        self.assertFalse(self.g.ok)
        self.assertIsNone(self.g.apicaller.response)
    
    def test_safemode(self):
        self.g.update_config({'GRIST_SAFEMODE': 'Y'})
        with self.assertRaises(GristApiInSafeMode):
            self.g.add_workspace('bogus')
        # safe mode leverages on dry run but should clean up after
        self.assertFalse(self.g.apicaller.dry_run)
        self.assertIsNone(self.g.apicaller.response)
        self.assertIsInstance(self.g.apicaller.request, PreparedRequest)

    def test_ok(self):
        try:
            st, res = self.g.see_workspace(1) # hopefully, not a valid ws id!
        except HTTPError:
            pass
        self.assertFalse(self.g.ok)
        st, res = self.g.see_team()
        self.assertTrue(self.g.ok)

    def test_custom_configurator_apicaller(self):
        # this has to be true, always
        self.assertIs(self.g.configurator, self.g.apicaller.configurator)
        # now let's try a custom configurator
        c = config.Configurator(config={'foo': 'bar'})
        g = api.GristApi(config={'baz':'42'}, custom_configurator=c)
        self.assertIs(self.g.configurator, self.g.apicaller.configurator)
        self.assertEqual(g.configurator.config['foo'], 'bar')
        self.assertEqual(g.configurator.config['baz'], '42')
        g.update_config({'foo':'foobar'})
        self.assertEqual(g.configurator.config['foo'], 'foobar')
        g.reconfig()
        self.assertNotIn('foo', g.configurator.config)
        # now let's try a custom api caller
        a = ApiCaller(configurator=c)
        g = api.GristApi(config={'baz':'42'}, custom_apicaller=a)
        self.assertIs(self.g.configurator, self.g.apicaller.configurator)
        # passing both should be impossible
        with self.assertRaises(GristApiNotConfigured):
            api.GristApi(custom_apicaller=a, custom_configurator=c)
        # changing configurator at runtime is possible but not supported
        new_c = config.Configurator()
        g.configurator = new_c
        g.apicaller.configurator = new_c # we need to sync manually
        self.assertNotIn('baz', g.configurator.config) # new clean configuration!

    def test_reconfig(self):
        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'N')
        # update_config is for incremental changes
        self.g.update_config({'GRIST_SAFEMODE': 'Y'})
        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'Y')
        self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'N')
        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'Y')
        # reconfig is for re-building from scratch
        self.g.reconfig({'GRIST_RAISE_ERROR': 'Y'})
        self.assertEqual(self.g.configurator.config['GRIST_RAISE_ERROR'], 'Y')
        self.assertEqual(self.g.configurator.config['GRIST_SAFEMODE'], 'N')

    def test_request_options(self):
        # as an example of extra-options, we test a timeout limit
        # let's make it so the server is 'http://10.255.255.1'
        self.g.update_config({'GRIST_SERVER_PROTOCOL': 'http://', 
                              'GRIST_TEAM_SITE': '10', 
                              'GRIST_API_SERVER': '255.255.1', 
                              # maybe we are self-managed instead
                              'GRIST_SELF_MANAGED_HOME': 'http://10.255.255.1'})
        # without this, test will take forever, before failing with ConnectTimeout
        self.g.apicaller.request_options = {'timeout': 1}
        with self.assertRaises(ConnectTimeout):
            st, res = self.g.see_team()

    def test_request_session(self):
        self.g.open_session()
        # simple GET api
        st, res = self.g.see_team()
        self.assertEqual(st, 200)
        name = str(time.time_ns())
        # POST
        st, ws_id = self.g.add_workspace(name, self.team_id)
        self.assertEqual(st, 200)
        self.assertIn('Cookie', self.g.apicaller.request.headers) # the session thing is working
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=ws_id)
        self.assertEqual(st, 200)
        # GET in download mode
        st, res = self.g.download_csv(name+'.csv', table_id='Table1',
                                      doc_id=doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # POST in upload mode
        f = HERE / 'imgtest.jpg'
        st, res = self.g.upload_attachments([f], doc_id=doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # PATCH
        name = str(time.time_ns())
        st, res = self.g.update_workspace(name, ws_id=ws_id)
        self.assertEqual(st, 200)
        # DELETE
        st, res = self.g.delete_doc(doc_id)
        self.assertEqual(st, 200)
        self.g.close_session()

class TestErrorContitions(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        cls.ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(cls.ws_id)
        
    def test_raise_error_before_call(self):
        # an example where Pygrister will fail even before posting the call
        with self.assertRaises(OSError):
            self.g.upload_attachments(['bogus_upload_file'], 
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsNone(self.g.apicaller.request)
        self.assertFalse(self.g.ok) # this is by design, see note in Apicall.ok

    def test_raise_error_bogus_url(self):
        # unresolvable url 
        self.g.update_config({'GRIST_SELF_MANAGED':'N', 
                              'GRIST_API_SERVER': 'bogus_server.com'})
        # note: GRIST_RAISE_ERROR setting won't make any difference here
        with self.assertRaises(RequestException):
            self.g.see_team()
        self.assertIsNone(self.g.apicaller.response)
        self.assertFalse(self.g.ok) # this is by design, see note in Apicall.ok
    
    def test_raise_error_no_server(self):
        # no Grist server there: this should be a simple 404...
        self.g.update_config({'GRIST_SELF_MANAGED':'N', 
                              'GRIST_TEAM_SITE':'www', 
                              'GRIST_API_SERVER':'example.com'})
        with self.assertRaises(HTTPError):
            self.g.see_team()
        self.g.update_config({'GRIST_RAISE_ERROR':'N'})
        st, res = self.g.see_team()
        self.assertEqual(st, 404)
    
    def test_raise_error_no_response(self):
        # an example where Grist won't return the call
        # (this is the same scenario as test_request_options)
        self.g.update_config({'GRIST_SERVER_PROTOCOL': 'http://', 
                              'GRIST_TEAM_SITE': '10', 
                              'GRIST_API_SERVER': '255.255.1', 
                              # if we are self-managed instead
                              'GRIST_SELF_MANAGED_HOME': 'http://10.255.255.1'})
        self.g.apicaller.request_options = {'timeout': 1}
        with self.assertRaises(ConnectTimeout):
            self.g.see_team()
        self.assertIsNone(self.g.apicaller.response)
        self.assertFalse(self.g.apicaller.ok) # this is by design, see note in Apicall.ok

    def test_raise_error_http401(self):
        # invalid api key will return http 401
        self.g.update_config({'GRIST_API_KEY': 'bogus_api_key'})
        with self.assertRaises(HTTPError):
            self.g.see_team()
        self.assertEqual(self.g.apicaller.response.status_code, 401)
        self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
        st, res = self.g.see_team()
        self.assertEqual(st, 401)
        # here Grist won't return regular json: we can convert though
        try:
            json.loads(self.g.apicaller.response_as_json())
        except json.JSONDecodeError:
            self.fail('Failed to convert a Http401 response to json.')

    def test_raise_error_http404(self):
        # bogus teams and docs result in http 404
        with self.assertRaises(HTTPError):
            self.g.see_team('bogus_team_id')
        with self.assertRaises(HTTPError):
            self.g.see_doc('bogus_doc_id')
        self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
        st, res = self.g.see_team('bogus_team_id')
        self.assertEqual(st, 404)
        st, res = self.g.see_doc('bogus_doc_id')
        self.assertEqual(st, 404)
    
    def test_no_raise_http_error(self):
        self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
        # many functions in Pygrister will patch the response body before 
        # returning it: when GRIST_RAISE_ERROR=N, this will happen also in 
        # case of Http error, so we must make sure that a failed post-processing 
        # will not result in spurious exceptions being raised. 
        # This is just a sample, not a full list of all problematic functions.
        _ = self.g.update_workspace('foo', 0)
        self.assertFalse(self.g.ok)
        _ = self.g.list_workspace_users(0)
        self.assertFalse(self.g.ok)
        _ = self.g.update_doc('foo', doc_id='bogus')
        self.assertFalse(self.g.ok)
        _ = self.g.list_doc_users(doc_id='bogus')
        self.assertFalse(self.g.ok)
        _ = self.g.list_tables('bogus_doc')
        self.assertFalse(self.g.ok)
        _ = self.g.list_cols('Table1', doc_id='bogus')
        self.assertFalse(self.g.ok)
        _ = self.g.list_attachments(doc_id='bogus')
        self.assertFalse(self.g.ok)
        _ = self.g.list_webhooks('bogus')
        self.assertFalse(self.g.ok)

class TestUsersNoScim(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']

    def test_delete_myself(self):
        # this is not yet implemented
        with self.assertRaises(api.GristApiNotImplemented):
            self.g.delete_myself('bogus')
    
    def test_enable_user(self):
        with self.assertRaises(HTTPError):
            self.g.enable_user(6666666666666)
        # let's hope low-ids like "4" are present by default
        # as we would need SCIM to create (or even list!) users
        self.g.update_config({'GRIST_RAISE_ERROR':'N'})
        st, res = self.g.enable_user(4, False)
        self.assertIn(st, (200, 400)) # http400 is for "user not found"
        if st == 200:
            self.assertIsNone(res)
        st, res = self.g.enable_user(4)
        self.assertIn(st, (200, 400)) # http400 is for "user not found"
        if st == 200:
            self.assertIsNone(res)
 
# note: to test scim ops, you need both GRIST_SELF_MANAGED here (because the 
# regular Grist SaaS does not support scim) and GRIST_ENABLE_SCIM on the server!
@unittest.skipIf(TEST_CONFIGURATION['GRIST_SELF_MANAGED'] == 'N', '')
@unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_SCIM_TESTS'] == 'N', '')
class TestUsersScim(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
    
    def test_see_user_and_myself(self):
        st, res = self.g.see_myself()
        self.assertEqual(st, 200)
        myself_id = int(res['id']) 
        st, res = self.g.see_user(myself_id)
        self.assertEqual(st, 200)

    def test_list_users_raw(self):
        # this demonstrates using the "raw" api call, with manual pagination
        st, res = self.g.list_users_raw(1, 10)
        self.assertEqual(st, 200)
        id_first = res['Resources'][0]['id']
        id_second = res['Resources'][1]['id']
        st, res = self.g.list_users_raw(2, 10)
        self.assertEqual(st, 200)
        self.assertEqual(res['Resources'][0]['id'], id_second)
        # passing an out-of-range index results in returning the first
        # items anyway: I believe this is a mistake...
        st, res = self.g.list_users_raw(1000000, 10)
        self.assertEqual(st, 200)
        # if this fails, it means the issue has been fixed:
        self.assertEqual(res['Resources'][0]['id'], id_first)
        # and now for the filters! I can't make this work... 
        # note that the very same filter will work in the search api...
        # if this fails, it means it was a bug and it has been fixed:
        with self.assertRaises(HTTPError):
            st, res = self.g.list_users_raw(filter='userName co example')

    def test_list_users(self):
        # this demonstates using the fancy, auto-paginating api call
        for st, res in self.g.list_users(): 
            self.assertEqual(st, 200)
        # now let's iterate manually
        user_list = self.g.list_users()
        self.assertEqual(len(user_list), 0)
        st, chunk = next(user_list)
        self.assertEqual(st, 200)
        self.assertGreater(len(user_list), 0)
        # here we can fix the out-of-range issue:
        user_list = self.g.list_users(1000000, 10)
        with self.assertRaises(StopIteration):
            next(user_list)

    @unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_USER_TESTS'] == 'N', '')
    def test_add_update_delete_user(self):
        base_name = str(time.time_ns())
        usr, email = f'{base_name}', f'{base_name}@example.com'
        st, user_id = self.g.add_user(usr, [email])
        self.assertEqual(st, 201)
        # let's change the locale for this user
        ops = [{'op': 'replace', 'path': 'locale', 'value': 'es'}]
        st, res = self.g.update_user(user_id, ops)
        self.assertEqual(st, 200)
        st, usr = self.g.see_user(user_id)
        self.assertEqual(st, 200)
        self.assertEqual(usr['locale'], 'es')
        # now delete the user
        st, res = self.g.delete_user(user_id)
        self.assertEqual(st, 204)
        self.assertIsNone(res)
        # now double-check that the user is gone
        self.g.update_config({'GRIST_RAISE_ERROR': 'N'})
        st, res = self.g.see_user(user_id)
        self.assertEqual(st, 404)

    def test_search_users_raw(self):
        # this demonstrates using the "raw" api call, with manual pagination
        st, res = self.g.search_users_raw(1, 10)
        self.assertEqual(st, 200)
        # and now for the filters... let's add a user first
        # (note: a bunch of "@getgrist" users should always be present)
        st, res = self.g.search_users_raw(filter='userName co getgrist')
        self.assertEqual(st, 200)

    def test_search_users(self):
        # this demonstates using the fancy, auto-paginating api call
        for st, res in self.g.search_users(): 
            self.assertEqual(st, 200)
        # now let's iterate manually
        user_list = self.g.search_users()
        self.assertEqual(len(user_list), 0)
        st, chunk = next(user_list)
        self.assertEqual(st, 200)
        self.assertGreater(len(user_list), 0)

    @unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_USER_TESTS'] == 'N', '')
    def test_bulk_users(self):
        # let's add a few users in bulk
        base_name = str(time.time_ns())
        users = [[f'{base_name}_01', f'{base_name}_01@example.com'],
                 [f'{base_name}_02', f'{base_name}_02@example.com'],
                 [f'{base_name}_03', f'{base_name}_03@example.com']]
        ops = [{'method': 'POST', 'path': '/scim/v2/Users', 
                'data': {'name': usr, 
                         'emails': {'value': eml, 'primary': True}}}
                for (usr, eml) in users]
        st, res = self.g.bulk_users(operations=ops)
        self.assertEqual(st, 200)
        # note: the api call should work, but each actual ops will fail with 
        # http 400, because the path is incorrect... can't figure out why, 
        # maybe a bug in the underlying Grist api
        self.assertEqual(res, [400, 400, 400]) # hoping this will fail some day!

    def test_see_scim_schemas(self):
        st, res = self.g.see_scim_schemas()
        self.assertEqual(st, 200)

    def test_see_scim_config(self):
        st, res = self.g.see_scim_config()
        self.assertEqual(st, 200)

    def test_see_scim_resources(self):
        st, res = self.g.see_scim_resources()
        self.assertEqual(st, 200)

class TestTeamSites(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
     
    def test_list_team_sites(self):
        st, res = self.g.list_team_sites()
        self.assertEqual(st, 200)
        self.assertIsInstance(res, list)
    
    def test_see_team(self):
        st, res = self.g.see_team(TestTeamSites.team_id)
        self.assertEqual(st, 200)
        self.assertIsInstance(res, dict)
        with self.assertRaises(HTTPError):
            st, res = self.g.see_team('bogus')

    def test_update_team(self):
        self.g.update_config({'GRIST_SAFEMODE': 'Y'})
        with self.assertRaises(api.GristApiInSafeMode):
            self.g.update_team('bogus', TestTeamSites.team_id)
        self.g.update_config({'GRIST_SAFEMODE': 'N'})
        st, res = self.g.update_team('apitestrenamed', TestTeamSites.team_id)
        self.assertEqual(st, 200)
        st, res = self.g.update_team('apitestteam', TestTeamSites.team_id)
        self.assertEqual(st, 200)
        self.assertIsNone(res)

    @unittest.skip
    def test_delete_team(self):
        # since there is no way to *create* a team in the first place, 
        # we won't test team deletion!
        pass

    def test_list_team_users(self):
        st, res = self.g.list_team_users(TestTeamSites.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    @unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_USER_TESTS'] == 'N', '')
    def test_update_team_users(self):
        name = str(time.time_ns())[-5:]
        users = {f'u{name}a@example.com': 'editors', 
                 f'u{name}b@example.com': 'owners'}
        st, res = self.g.update_team_users(users, self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

class TestWorkspaces(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        gristapi = api.GristApi(config=TEST_CONFIGURATION)
        st, res = gristapi.list_workspaces(cls.team_id) # should work :)
        cls.workspace_id = res[0]['id']
        cls.workspace_name = res[0]['name']
        total_apicalls.append(gristapi.apicaller.apicalls)

    def test_list_workspaces(self):
        st, res = self.g.list_workspaces(self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    def test_see_workspace(self):
        st, res = self.g.see_workspace(self.workspace_id)
        self.assertIsInstance(res, dict)
        self.assertEqual(st, 200)

    def test_add_delete_workspace(self):
        self.g.update_config({'GRIST_SAFEMODE': 'Y'})
        name = str(time.time_ns())
        with self.assertRaises(api.GristApiInSafeMode):
            self.g.add_workspace(name, self.team_id)
        self.g.update_config({'GRIST_SAFEMODE': 'N'})
        name = str(time.time_ns())
        st, ws_id = self.g.add_workspace(name, self.team_id)
        self.assertEqual(st, 200)
        st, res = self.g.delete_workspace(ws_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        
    def test_update_workspace(self):
        name = str(time.time_ns())
        st, res = self.g.update_workspace(name, self.workspace_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.update_workspace(self.workspace_name, self.workspace_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_list_workspace_users(self):
        st, res = self.g.list_workspace_users(self.workspace_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
    
    @unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_USER_TESTS'] == 'N', '')
    def test_update_workspace_users(self):
        name = str(time.time_ns())[-5:]
        users = {f'u{name}a@example.com': 'editors', 
                 f'u{name}b@example.com': 'owners'}
        # note: must be added as team users first!
        st, res = self.g.update_team_users(users, self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.update_workspace_users(users, self.workspace_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_workspace_cross_site(self):
        # see/add workspace in our site "from" another site
        self.g.update_config({'GRIST_TEAM_SITE': 'docs'})
        st, res = self.g.see_workspace(self.workspace_id)
        self.assertIsInstance(res, dict)
        self.assertEqual(st, 200)
        name = str(time.time_ns())
        st, res = self.g.add_workspace(name, self.team_id)
        self.assertIsInstance(res, int)
        self.assertEqual(st, 200)

class TestDocs(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        cls.workspace_id, cls.workspace_name = _make_ws(cls.team_id)

    def test_create_delete_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.delete_doc(doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        # shouldn't be possible in safemode
        self.g.update_config({'GRIST_SAFEMODE': 'Y'})
        name = str(time.time_ns())
        with self.assertRaises(api.GristApiInSafeMode):
            self.g.add_doc(name, self.workspace_id)

    def test_delete_doc_history(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.delete_doc_history(0, doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    @unittest.expectedFailure
    def test_create_delete_doc_cross_site(self):
        # fun fact: cross-site doc creation is allowed...
        self.g.update_config({'GRIST_TEAM_SITE': 'docs'})
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        # ...but deletion is not, EXCEPT when we operate from 'docs', apparently 
        # this should be investigated further, maybe 'docs' is a catch-all 
        # team id hosting "all" the docs? In the meantime, we mark this one 
        # as an expectedFailure
        with self.assertRaises(HTTPError):
            self.g.delete_doc(doc_id)

    def test_see_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.see_doc(doc_id, self.team_id)
        self.assertIsInstance(res, dict)
        self.assertEqual(st, 200)

    def test_update_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.update_doc('new'+name, doc_id=doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.update_doc(pinned=True, doc_id=doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_move_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, ws_id = self.g.add_workspace('ws'+name, self.team_id)
        self.assertIsInstance(ws_id, int)
        self.assertEqual(st, 200)
        st, res = self.g.move_doc(ws_id, doc_id, self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_copy_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, ws_id = self.g.add_workspace('ws'+name, self.team_id)
        self.assertIsInstance(ws_id, int)
        self.assertEqual(st, 200)
        st, res = self.g.copy_doc(ws_id, 'new'+name, True, doc_id, self.team_id)
        self.assertEqual(st, 200)

    def test_reload_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.reload_doc(doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
    
    def test_enable_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.enable_doc(False, doc_id=doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.enable_doc(True, doc_id=doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_recovery_doc(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.set_recovery_mode(True, doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.set_recovery_mode(False, doc_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_list_doc_users(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.list_doc_users(doc_id, self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    @unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_USER_TESTS'] == 'N', '')
    def test_update_doc_users(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        users = {f'u{name[-5:]}a@example.com': 'editors', 
                 f'u{name[-5:]}b@example.com': 'owners'}
        # note: no need to add these to your team/workspace first
        st, res = self.g.update_doc_users(users, doc_id=doc_id, 
                                          team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_download_upload_sqlite(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.download_sqlite(Path(f'{name}.grist'), 
            nohistory=True, template=True, doc_id=doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        # now we re-import the same file in the same ws with a different name
        # note: extension has to be ".grist"!
        st, res = self.g.upload_sqlite(Path(f'{name}.grist'),
                            f'imported_{name}', ws_id=self.workspace_id)
        self.assertEqual(st, 200)
        self.assertIsInstance(res, str)

    def test_download_excel(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.download_excel(Path(f'{name}.xls'), table_id='Table1',
                                        doc_id=doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
    
    def test_download_csv(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.download_csv(Path(f'{name}.csv'), table_id='Table1',
                                      doc_id=doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_download_schema(self):
        name = str(time.time_ns())
        st, doc_id = self.g.add_doc(name, ws_id=self.workspace_id)
        self.assertIsInstance(doc_id, str)
        self.assertEqual(st, 200)
        st, res = self.g.download_schema('Table1', doc_id=doc_id, 
                                          team_id=self.team_id)
        self.assertIsInstance(res, dict)
        self.assertEqual(st, 200)
        st, res = self.g.download_schema('Table1', filename=Path(f'{name}.json'), 
                                         doc_id=doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

class TestRecordAccess(BaseTestPyGrister): 
    # we test "/records", "/data" and "/sql" endpoints here (no converters)
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(ws_id)
        cols = [{'id': 'Astr', 'fields': {'label': 'Astr', 'type': 'Text'}},
                {'id': 'Bnum', 'fields': {'label': 'Bnum', 'type': 'Numeric'}},
                {'id': 'Cint', 'fields': {'label': 'Cint', 'type': 'Int'}},
                {'id': 'Dbol', 'fields': {'label': 'Dbol', 'type': 'Bool'}},]
        tables = [{'id': 'T'+name, 'columns': cols}]
        gristapi = api.GristApi(config=TEST_CONFIGURATION)
        st, res = gristapi.add_tables(tables, cls.doc_id, cls.team_id)
        cls.table_id = res[0]
        total_apicalls.append(gristapi.apicaller.apicalls)
    
    def test_list_records(self):
        records = [{'Astr': 'test 1', 'Bnum': 1.1, 'Cint': 1, 'Dbol': True},
                   {'Astr': 'test 2', 'Bnum': 2.2, 'Cint': 2, 'Dbol': False}, 
                   {'Astr': 'test 3', 'Bnum': 3.3, 'Cint': 3, 'Dbol': False},
                   {'Astr': 'test 4', 'Bnum': 4.4, 'Cint': 4, 'Dbol': True},]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 4)
        self.assertEqual(st, 200)
        st, res = self.g.list_records(self.table_id, 
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        st, res = self.g.list_records(self.table_id, sort='-Cint', limit=2, 
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
    
    def test_list_records_with_filter(self):
        st, res = self.g.list_records(self.table_id, filter={'Astr': ['test 3']}, 
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    def test_add_and_update_records(self):
        records = [{'Astr': 'test1', 'Bnum': 1.1, 'Cint': 1, 'Dbol': True},
                   {'Astr': 'test2', 'Bnum': 2.2, 'Cint': 2, 'Dbol': False}]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        records = [{'id': 1, 'Astr': 'modified', 'Bnum': 1.12},
                   {         'Astr': 'mod!!', 'Bnum': 2.22, 'id': 2}]
        st, res = self.g.update_records(self.table_id, records, 
                                        doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
    
    @unittest.skip  # not really our fault, I guess
    def test_add_records_noparse(self):
        # the "noparse" param is not enforced?
        records = [{'Cint': 'not a number'}]
        st, res = self.g.add_records(self.table_id, records, noparse=False,
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        st, res = self.g.add_records(self.table_id, records, noparse=True,
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    def test_add_update_records(self):
        records = [{'Astr': 'toupdate1', 'Bnum': 5.1, 'Cint': 5, 'Dbol': True},
                   {'Astr': 'toupdate2', 'Bnum': 6.2, 'Cint': 6, 'Dbol': False}]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        records = [{'require': {'Astr': 'toupdate1'}, 'fields': {'Cint': 55}}, 
                   {'require': {'Astr': 'toupdate2'}, 'fields': {'Bnum': 6.3}},
                   {'require': {'Astr': 'new rec'}, 'fields': {'Dbol': True}}]
        st, res = self.g.add_update_records(self.table_id, 
                        records, doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_data_delete(self): # the only non-deprecated /data endpoint
        records = [{'Astr': 'test1', 'Bnum': 1.1, 'Cint': 1, 'Dbol': True},
                   {'Astr': 'test2', 'Bnum': 2.2, 'Cint': 2, 'Dbol': False}, 
                   {'Astr': 'test3', 'Bnum': 3.3, 'Cint': 3, 'Dbol': False}]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        st, res = self.g.delete_rows(self.table_id, [1, 2], doc_id=self.doc_id, 
                                     team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_sql_and_sql_with_params(self): 
        records = [{'Astr': 'test sql1', 'Bnum': 1.1, 'Cint': 1, 'Dbol': True},
                   {'Astr': 'test sql2', 'Bnum': 2.2, 'Cint': 2, 'Dbol': False}, 
                   {'Astr': 'test sql3', 'Bnum': 3.3, 'Cint': 3, 'Dbol': False}]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        sql = f'select * from {self.table_id}'  # no trailing ";" !
        st, res = self.g.run_sql(sql, doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        sql = f'select * from {self.table_id} where Cint>?'
        qargs = [1]
        st, res = self.g.run_sql_with_args(sql, qargs, doc_id=self.doc_id, 
                                           team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

class TestConverters(BaseTestPyGrister): 
    # we test converters for "/records" and "/sql" endpoints here
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(ws_id)
        cols = [{'id': 'Test', 'fields': {'label': 'Test', 'type': 'Text'}},
                {'id': 'Sort', 'fields': {'label': 'Sort', 'type': 'Text'}},
                {'id': 'Bnum', 'fields': {'label': 'Bnum', 'type': 'Numeric'}},
                {'id': 'Cint', 'fields': {'label': 'Cint', 'type': 'Int'}},
                {'id': 'Dbol', 'fields': {'label': 'Dbol', 'type': 'Bool'}},
                {'id': 'Edate', 'fields': {'label': 'Edate', 'type': 'Date'}},]
        tables = [{'id': 'T'+name, 'columns': cols}]
        gristapi = api.GristApi(config=TEST_CONFIGURATION)
        st, res = gristapi.add_tables(tables, cls.doc_id, cls.team_id)
        cls.table_id = res[0]
        total_apicalls.append(gristapi.apicaller.apicalls)

    def test_list_records(self):
        records = [{'Test': 'list_records', 'Sort': 'a', 
                    'Bnum': 1.1, 'Edate': 1602280800},
                   {'Test': 'list_records', 'Sort': 'b', 
                    'Bnum': 2, 'Edate': None}, 
                   {'Test': 'list_records', 'Sort': 'c', 
                    'Bnum': 'hello', 'Edate': 'hello'},]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # first we try without converters...
        st, res = self.g.list_records(self.table_id, 
                                      filter={'Test': ['list_records']}, sort='Sort',
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertEqual(res[0]['Bnum'], 1.1)
        self.assertEqual(res[1]['Bnum'], 2)
        self.assertEqual(res[2]['Bnum'], 'hello')
        self.assertEqual(res[0]['Edate'], 1602280800)
        self.assertEqual(res[1]['Edate'], None)
        self.assertEqual(res[2]['Edate'], 'hello')
        # now let's try converters:
        conv = {self.table_id: 
                    {'Bnum': lambda i: float(i),
                     'Edate': lambda i: datetime.fromtimestamp(i),}  # (1)
                }
        self.g.out_converter = conv
        st, res = self.g.list_records(self.table_id, 
                                      filter={'Test': ['list_records']}, sort='Sort',
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertEqual(res[0]['Bnum'], 1.1)
        self.assertEqual(res[1]['Bnum'], 2.0)
        self.assertEqual(res[2]['Bnum'], 'hello')
        self.assertEqual(res[0]['Edate'], datetime(2020, 10, 10))
        self.assertEqual(res[1]['Edate'], None)
        self.assertEqual(res[2]['Edate'], 'hello')

        # (1) note: this works here, but it's very naive: write timezone-aware 
        # converters in your real code! The Grist Api will silently compensate 
        # for your document (local) timezone when inserting/retrieving dates, 
        # so what ends up stored in the database will be different from the naive 
        # date that you have passed to the Api. For instance, at the time of this 
        # writing my Grist document was set at UTC+2, so while I passed 
        # timestamp "1602280800", the number stored in the database was 
        # "1602201600" (2 hours earlier) and, when I checked with the Grist GUI, 
        # the date diplayed was indeed "2020-10-09" instead of "2020-10-10" 
        # as I intended. 
        # See https://support.getgrist.com/dates/#time-zones and 
        # https://xkcd.com/1883/

    def test_add_records(self):
        conv = {self.table_id: 
                    {'Bnum': lambda i: float(i),
                     'Edate': lambda i: int(datetime.timestamp(i)),}  # (1) above
                }
        self.g.in_converter = conv
        # we add records with a converter
        records = [{'Test': 'add_records', 'Sort': 'a', 
                    'Bnum': 1.1, 'Edate': datetime(2020, 10, 10)},
                   {'Test': 'add_records', 'Sort': 'b', 
                    'Bnum': '2', 'Edate': datetime(2020, 10, 10)},]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # now we retrieve them without any "out" converter set
        st, res = self.g.list_records(self.table_id, 
                                      filter={'Test': ['add_records']}, sort='Sort',
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertEqual(res[0]['Bnum'], 1.1)
        self.assertEqual(res[1]['Bnum'], 2.0)
        self.assertEqual(res[1]['Edate'], 1602280800)

    def test_add_records_weak_converter(self):
        # note: Pygrister will fix Value/TypeErrors for you for "outgoing"
        # converters only (ie, when retrieving data), 
        # but NOT for "ingoing" converters (when uploading data)
        conv = {self.table_id: {'Bnum': lambda i: float(i)}}
        self.g.in_converter = conv
        # this is a "weak" converter that ignores possible errors
        with self.assertRaises(ValueError):
            st, res = self.g.add_records(self.table_id, [{'Bnum': 'hello'}], 
                                         doc_id=self.doc_id, team_id=self.team_id)
        with self.assertRaises(TypeError):
            st, res = self.g.add_records(self.table_id, [{'Bnum': None}], 
                                         doc_id=self.doc_id, team_id=self.team_id)

    def test_add_update_records(self):
        records = [{'Test': 'add_update', 'Sort': 'to_update_a', 'Bnum': 1.1},
                   {'Test': 'add_update', 'Sort': 'to_update_b', 'Bnum': 2}]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # let's add a converter...
        conv = {self.table_id: 
                    {'Bnum': lambda i: float(i),
                     'Edate': lambda i: int(datetime.timestamp(i)),} 
                }
        self.g.in_converter = conv
        records = [{'require': {'Sort': 'to_update_a'}, 
                    'fields': {'Bnum': '5'}}, 
                   {'require': {'Sort': 'to_update_b'}, 
                    'fields': {'Edate': datetime(2020, 10, 10)}},]
        st, res = self.g.add_update_records(self.table_id, 
                        records, doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # now let's retrieve the records without a converter
        st, res = self.g.list_records(self.table_id, 
                                      filter={'Test': ['add_update']}, sort='Sort',
                                      doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertEqual(res[0]['Bnum'], 5.0)
        self.assertEqual(res[1]['Edate'], 1602280800)

    def test_sql(self):
        records = [{'Test': 'test_sql', 'Sort': 'a', 
                    'Bnum': 1.1, 'Edate': 1602280800},
                   {'Test': 'test_sql', 'Sort': 'b', 
                    'Bnum': 2, 'Edate': None}, 
                   {'Test': 'test_sql', 'Sort': 'c', 
                    'Bnum': 'hello', 'Edate': 'hello'},]
        st, res = self.g.add_records(self.table_id, records, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertEqual(st, 200)
        # now we add an "out" converter:
        conv = {'sql': 
                    {'Bnum': lambda i: float(i),
                     'Edate': lambda i: datetime.fromtimestamp(i),}
                }
        self.g.out_converter = conv
        sql = f'select * from {self.table_id} where Test=? order by Sort' 
        qargs = ['test_sql']
        st, res = self.g.run_sql_with_args(sql, qargs, doc_id=self.doc_id, 
                                           team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertEqual(res[0]['Bnum'], 1.1)
        self.assertEqual(res[1]['Bnum'], 2.0)
        self.assertEqual(res[2]['Bnum'], 'hello')
        self.assertEqual(res[0]['Edate'], datetime(2020, 10, 10))
        self.assertEqual(res[1]['Edate'], None)
        self.assertEqual(res[2]['Edate'], 'hello')

class TestTables(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        cls.ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(cls.ws_id)

    def test_list_tables(self):
        st, res = self.g.list_tables(self.doc_id, self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
    
    def test_add_tables(self):
        name = 'T'+str(time.time_ns())
        tables = [{'id': name, 'columns': 
                   [{'id': 'col1', 'fields': {'label': 'Col 1'}}]}]
        st, res = self.g.add_tables(tables, self.doc_id, self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    @unittest.expectedFailure
    def test_update_tables(self):
        name = 'T'+str(time.time_ns())
        tables = [{'id': name, 'columns': 
                   [{'id': 'col1', 'fields': {'label': 'Col 1'}}]}]
        st, res = self.g.add_tables(tables, self.doc_id, self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        tables = [{'id': name, 'columns': 
                   [{'id': 'col1', 'fields': {'label': 'Col 1bis'}}]}]
        # now, this *looks like* the same innocent payload as before, with 
        # only a minor edit... but this time, boom! the Grist api will go 
        # full http400 with a rather cryptic message like 
        # 'error': 'Invalid payload', 'details': 
        # {'userError': 'Error: body.tables[0] is not a RecordWithStringId; 
        #   body.tables[0].fields is missing'}
        # This will need a few more rounds of trial/error I'm afraid...
        st, res = self.g.update_tables(tables, self.doc_id, self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

class TestCols(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(ws_id)
        cls.table_id = 'Table1' # we trust this to be always present in new docs

    def test_list_cols(self):
        st, res = self.g.list_cols(self.table_id, doc_id=self.doc_id, 
                                   team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
    
    def test_add_update_delete_cols(self):
        name = 'col'+str(time.time_ns())[-5:]
        cols = [{'id': name+'_a', 'fields': {'label': name+'_a'}},
                {'id': name+'_b', 'fields': {'label': name+'_b',
                                             'type': 'Int'}}]
        st, res = self.g.add_cols(self.table_id, cols=cols, 
                                  doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 2)
        self.assertEqual(st, 200)
        cols = [{'id': name+'_a', 'fields': {'label': name+'_newa'}},
                {'id': name+'_b', 'fields': {'label': name+'_newb'}}]
        st, res = self.g.update_cols(self.table_id, cols=cols, 
                                     doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.delete_column(self.table_id, name+'_newa', 
                                       doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
    
    def test_add_cols_with_options(self):
        name = 'col'+str(time.time_ns())[-5:]
        # this is the example from the Grist api console
        cols = [{'id': name+'_opts', 
                 'fields': {
                    'label': name+'_opts',
                    'type': 'Choice', 
                    'widgetOptions': {
                        'choices': ['New', 'Old'],
                        'choiceOptions': {
                            'New': {
                                'fillColor': '#FF0000',
                                'textColor': '#FFFFFF'}}}}}]
        st, res = self.g.add_cols(self.table_id, cols=cols, 
                                  doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    def test_add_update_cols(self):
        name = 'col'+str(time.time_ns())[-5:]
        cols = [{'id': name+'_a', 'fields': {'label': name+'_a'}},
                {'id': name+'_b', 'fields': {'label': name+'_b',
                                             'type': 'Int'}}]
        st, res = self.g.add_cols(self.table_id, cols=cols, doc_id=self.doc_id, 
                                  team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        cols = [{'id': name+'_a', 'fields': {'label': name+'_newa'}}, # updated
                {'id': name+'_c', 'fields': {'label': name+'_c'}}]    # added
        st, res = self.g.add_update_cols(self.table_id, cols=cols, noadd=False, 
                                         noupdate=False, doc_id=self.doc_id, 
                                         team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

class TestAttachments(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        cls.ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(cls.ws_id)

    def test_list_attachments(self):
        st, res = self.g.list_attachments(doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
    
    def test_list_attachment_with_filter(self):
        filter = {'fileName': ['cat', 'dog']}
        st, res = self.g.list_attachments(filter=filter, doc_id=self.doc_id, 
                                          team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    def test_upload_download_attachments(self):
        f = HERE / 'imgtest.jpg'
        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
                                            team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 1)
        self.assertEqual(st, 200)
        # upload multiple attachments
        st, res = self.g.upload_attachments([f, f, f], doc_id=self.doc_id, 
                                            team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(len(res), 3)
        self.assertEqual(st, 200)
        name = Path(f'att_{time.time_ns()}.jpg')
        st, res = self.g.download_attachment(name, 1, doc_id=self.doc_id, 
                                             team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
    
    def test_download_restore_attachments(self):
        # this is for the download/upload "archive" apis
        f = HERE / 'imgtest.jpg'
        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
                                            team_id=self.team_id)
        self.assertEqual(st, 200)
        name = Path(f'downloaded_atts_{time.time_ns()}.tar')
        st, res = self.g.download_attachments(name, doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertIsNone(res)
        self.assertTrue(name.is_file())
        # now we re-upload the tar file
        st, res = self.g.upload_restore_attachments(name, doc_id=self.doc_id, 
                                                    team_id=self.team_id)
        self.assertEqual(st, 200)
        self.assertDictEqual(res, {'added': 0, 'errored': 0, 'unused': 1})
        # downloading should work even if no attachments are present
        doc_id = _make_doc(self.ws_id)
        st, res = self.g.download_attachments(doc_id=doc_id, 
                                              team_id=self.team_id)
        self.assertEqual(st, 200)

    def test_see_attachment(self):
        # make sure we have at least one attachment to see
        f = HERE / 'imgtest.jpg'
        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
                                            team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        # note that id=1 could not be the one we just uploaded... 
        st, res = self.g.see_attachment(1, doc_id=self.doc_id, 
                                        team_id=self.team_id)
        self.assertIsInstance(res, dict)
        self.assertEqual(st, 200)

    def test_purge_attachments(self):
        # upload a new attachment, which is not referenced in any table - unused!
        f = HERE / 'imgtest.jpg'
        st, res = self.g.upload_attachments([f], doc_id=self.doc_id, 
                                            team_id=self.team_id)
        self.assertEqual(st, 200)
        st, res = self.g.list_attachments(doc_id=self.doc_id, 
                                          team_id=self.team_id)
        self.assertEqual(st, 200)
        att_num = len(res)
        assert att_num >= 1 # there should be at least one attachment now
        st, res = self.g.delete_unused_attachments(doc_id=self.doc_id, 
                                                   team_id=self.team_id)
        self.assertEqual(st, 200)
        st, res = self.g.list_attachments(doc_id=self.doc_id, 
                                          team_id=self.team_id)
        self.assertEqual(st, 200)
        assert len(res) < att_num # at least one attachment should be gone
        
# to test external attachments, you must run a self-hosted instance of Grist 
# with "GRIST_EXTERNAL_ATTACHMENTS_MODE=snapshots", and a S3-compatible bucket 
@unittest.skipIf(TEST_CONFIGURATION['GRIST_TEST_RUN_EXT_ATTACH_TESTS'] == 'N', '')
class TestExternalAttachments(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(ws_id)

    def test_store(self):
        # this is for the various "store" apis (see, update, settings)
        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertEqual(res, 'internal')
        st, res = self.g.update_attachment_store(internal_store=False, 
                            doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsNone(res)
        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertEqual(res, 'external')
        st, res = self.g.list_store_settings(doc_id=self.doc_id, 
                                             team_id=self.team_id)
        self.assertEqual(st, 200) # not really much to verify here... 

    def test_transfer(self):
        # this is for the transfering files apis, between external/internal store
        st, res = self.g.update_attachment_store(doc_id=self.doc_id, 
                                                 team_id=self.team_id)
        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertEqual(res, 'internal')
        f = HERE / 'imgtest.jpg'
        st, res = self.g.upload_attachments([f, f, f], doc_id=self.doc_id, 
                                            team_id=self.team_id)
        st, res = self.g.update_attachment_store(internal_store=False, 
                            doc_id=self.doc_id, team_id=self.team_id)
        st, res = self.g.see_attachment_store(doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertEqual(res, 'external')
        # now we should have some attachments to transfer
        st, res = self.g.transfer_attachments(doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertTrue(res['status']['isRunning'])
        # "see_transfer_status" return depends on when we call it... 
        st, res = self.g.see_transfer_status(doc_id=self.doc_id, 
                                             team_id=self.team_id)
        self.assertEqual(st, 200) # little we can test


class TestWebhooks(BaseTestPyGrister):
    @classmethod
    def setUpClass(cls):
        cls.team_id = TEST_CONFIGURATION['GRIST_TEAM_SITE']
        ws_id, name = _make_ws(cls.team_id)
        cls.doc_id = _make_doc(ws_id)
        cls.table_id = 'Table1' # we trust this to be always present in new docs

    def test_list_webhooks(self):
        st, res = self.g.list_webhooks(doc_id=self.doc_id, team_id=self.team_id)
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)

    def test_add_update_delete_webhooks(self):
        name = 'wh'+str(time.time_ns())
        wh = {'fields': {'name': name, 'memo': 'memo!', 
              'url': 'https://www.example.com',
              'enabled': True, 'eventTypes': ['add'], 'isReadyColumn': None, 
              'tableId': 'Table1'}}
        st, res = self.g.add_webhooks(webhooks=[wh], doc_id=self.doc_id, 
                                      team_id=self.team_id)
        if st == 403 and TEST_CONFIGURATION['GRIST_SELF_MANAGED'] == 'Y':
            # a special-case failure: if you are self-hosting, you should 
            # have an ALLOWED_WEBHOOK_DOMAINS server-side env variable here; 
            # you may even set it to "*", see Grist docs for details
            self.fail('You must add "example.com" to your ALLOWED_WEBHOOK_DOMAINS.')
        self.assertIsInstance(res, list)
        self.assertEqual(st, 200)
        wh_id = res[0]
        wh['fields']['memo'] = 'memo updated!'
        st, res = self.g.update_webhook(wh_id, wh, doc_id=self.doc_id, 
                                        team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)
        st, res = self.g.delete_webhook(wh_id, doc_id=self.doc_id, 
                                        team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)

    def test_empty_payloads_queue(self):
        st, res = self.g.empty_payloads_queue(doc_id=self.doc_id, 
                                              team_id=self.team_id)
        self.assertIsNone(res)
        self.assertEqual(st, 200)


if __name__ == '__main__':
    unittest.main()



"""
Pygrister exception hierarchy. 
------------------------------

Exceptions listed here can be raised by Pygrister, and they concern 
its internal functioning. Note that Grist API calls themselves, however, 
are delegated to Requests, and will (much more frequently!) throw 
the exceptions provided by Requests.
"""

class GristApiException(Exception): 
    """The base GristApi exception."""
    pass

class GristApiNotConfigured(GristApiException): 
    """A configuration error occurred."""
    pass

class GristApiNotImplemented(GristApiException): 
    "This API is not yet implemented by Pygrister."""
    pass

class GristApiInSafeMode(GristApiException): 
    """Pygrister is in safe mode, no writing to the db is possible."""
    pass



import os, os.path
import json as modjson
from pprint import pformat

from pygrister.exceptions import GristApiNotConfigured

# This is the default Pygrister configuration.
# Use only non-empty strings as config values.

PYGRISTER_CONFIG = {
    'GRIST_API_KEY': '<your_api_key_here>',
    'GRIST_SELF_MANAGED': 'N',
    'GRIST_SELF_MANAGED_HOME': 'http://localhost:8484',
    'GRIST_SELF_MANAGED_SINGLE_ORG': 'Y',
    'GRIST_SERVER_PROTOCOL': 'https://',
    'GRIST_API_SERVER': 'getgrist.com',
    'GRIST_API_ROOT': 'api',
    'GRIST_TEAM_SITE': 'docs',
    'GRIST_WORKSPACE_ID': '0', # this should be a string castable to int
    'GRIST_DOC_ID': '<your_doc_id_here>',
    'GRIST_RAISE_ERROR': 'Y',
    'GRIST_SAFEMODE': 'N',
}

def apikey2output(apikey: str) -> str:
    """Obfuscate the secret Grist API key for output printing."""
    klen = len(apikey)
    return apikey if klen < 5 else f'{apikey[:2]}<{klen-4}>{apikey[-2:]}'


class Configurator:
    def __init__(self, config: dict[str, str]|None = None):
        self.config = dict()     # the actual, current configuration
        self.server = ''         # the current api server url
        self.raise_option = True # if we should raise Http errors
        self.safemode = False    # read-only mode
        self.reconfig(config)

    @staticmethod
    def get_config() -> dict[str, str]:
        """Return the Pygrister global configuration dictionary. 
        
        This is the "static" configuration setup, not counting anything 
        you may alter at runtime.
        Config keys are first searched in ``config.py``, then in 
        ``~/.gristapi/config.json``, and finally in matching env variables. 
        See ``config.py`` for a list of the config keys currently in use.
        """
        config = dict(PYGRISTER_CONFIG)
        pth = os.path.join(os.path.expanduser('~'), '.gristapi/config.json')
        if os.path.isfile(pth):
            with open(pth, 'r') as f:
                config.update(modjson.loads(f.read()))
        for k in config.keys():
            try:
                config[k] = os.environ[k]
            except KeyError:
                pass
        return config

    @staticmethod
    def config2output(config: dict[str, str], multiline: bool = False) -> str:
        """Format the Pygrister configuration as a string for output printing."""
        if not config: 
            return '{<empty>}'
        cfcopy = dict(config)
        cfcopy['GRIST_API_KEY'] = apikey2output(cfcopy.get('GRIST_API_KEY', ''))
        return pformat(cfcopy) if multiline else str(cfcopy)

    def reconfig(self, config: dict[str, str]|None = None) -> None:
        """Reload the configuration options. 
        
        Call this function if you have just updated config files/env. vars 
        at runtime, and/or pass a dictionary to the ``config`` parameter 
        to override existing config keys for the time being, eg.::

            grist.reconfig({'GRIST_TEAM_SITE': 'newteam'})

        now all future api calls will be directed to the new team site. 

        Note: this will re-build your configuration from scratch, then appy 
        the ``config`` parameter on top. To edit your *existent* configuration 
        instead, use the ``update_config`` function.
        """
        self.config = self.get_config()
        if config is not None:
            self.config.update(config)
        self._post_reconfig()

    def update_config(self, config: dict[str, str]) -> None:
        """Edit the configuration options.
        
        Call this function to edit your current runtime configuration: 
        pass a dictionary to the ``config`` parameter to override existing 
        config keys for the time being, eg.::

            grist.reconfig({'GRIST_TEAM_SITE': 'newteam'})

        now all future api calls will be directed to the new team site. 

        Note: this will apply the ``config`` parameter on top of your 
        existing configuration. To re-build the configuration from scratch, 
        use the ``reconfig`` function instead.
        """
        self.config.update(config)
        self._post_reconfig()

    def _post_reconfig(self): # check and cleanup after config is changed
        if not self.config or not all(self.config.values()):
            msg = f'Missing config values.\n{self.config2output(self.config)}'
            raise GristApiNotConfigured(msg)
        ws_id = self.config['GRIST_WORKSPACE_ID']
        try:
            _ = int(ws_id)
        except ValueError:
            msg = f'Workspace ID must be castable to integer, not "{ws_id}".'
            raise GristApiNotConfigured(msg)
        self.server = self.make_server()
        self.raise_option = (self.config['GRIST_RAISE_ERROR'] == 'Y')
        self.safemode = (self.config['GRIST_SAFEMODE'] == 'Y')

    def make_server(self, team_name: str = '') -> str:
        """Construct the "server" part of the API url, up to "/api". 
        
        A few options are possible, depending on the type of Grist hosting.
        The only moving part, as far as the GristApi class is concerned, 
        is the team name.
        """
        cf = self.config
        the_team = team_name or cf['GRIST_TEAM_SITE']
        if cf['GRIST_SELF_MANAGED'] == 'N':
            # the usual SaaS Grist: "https://myteam.getgrist.com/api"
            return f'{cf["GRIST_SERVER_PROTOCOL"]}{the_team}.' + \
                f'{cf["GRIST_API_SERVER"]}/{cf["GRIST_API_ROOT"]}'
        else:
            if cf['GRIST_SELF_MANAGED_SINGLE_ORG'] == 'Y':
                # self-managed, mono-team: "https://mygrist.com/api"
                return f'{cf["GRIST_SELF_MANAGED_HOME"]}/{cf["GRIST_API_ROOT"]}'
            else:
                # self-managed: "https://mygrist.com/o/myteam/api"
                return f'{cf["GRIST_SELF_MANAGED_HOME"]}/o/{the_team}' + \
                    f'/{cf["GRIST_API_ROOT"]}'

    def select_params(self, doc_id: str = '', team_id: str = ''):
        doc = doc_id or self.config['GRIST_DOC_ID']
        if not team_id:
            server = self.server
        else:
            server = self.make_server(team_name=team_id)
        return doc, server
